
// Complex C++ example for NumGRID project
// This programs tests all supported MPI functions

/*
  :    ( MPI)   .  
    .   ,  ,  .
       Grid.    ,      
      .    .

    .

  TOTAL_STEPS .   :
 1)  random(POINT_RAND) * size / 2   -.  2     
      MPI_Send/Recv ( 50%)  MPI_Sendrecv(50%).
 2)  random(GROUP_RAND)   OP_BCAST,OP_REDUCE,OP_ALLREDUCE, OP_ALLGATHER ( )
 3.    (MPI)         ( MAX_DELAY_TIME )

   .       MPI_Float  DATA_SIZE .

*/

//bool GenerateNewMap = false; 
#include "gmpi.h"

bool GenerateNewMap = true; 
//#include <mpi.h>


#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <windows.h>
#include <vector>
#include <string>


#define MAX_PROCESSES 20
#define MAX_DELAY_TIME 20
#define POINT_RAND  10
#define GROUP_RAND  10

#define TOTAL_STEPS  15  //    

#define DATA_SIZE 5 //      

//  MPI 
#define OP_EMPTY     -1
#define OP_SEND      0
#define OP_RECV      1
#define OP_SENDRECV  2

#define OP_REDUCE    3 
#define OP_ALLREDUCE 4
#define OP_BCAST     5
#define OP_ALLGATHER 6

struct Operation
{
  int code;
  int destRank;
  MPI_Datatype opType;
  int delay;
};

std::vector <Operation> queue;

std::vector <Operation> map[MAX_PROCESSES];

#define MAX_STEPS 5


float data[DATA_SIZE], data1[DATA_SIZE];
//float data2[DATA_SIZE * MAX_PROCESSES];
float data2[DATA_SIZE * MAX_PROCESSES];


////////////////////////////////////////////////////////////////////////////
void writeResults(int rank)
{
  std::string str = "";
  int i;
  char buf[100];
  
  str += "Rank = ";
  itoa(rank, buf,10); str +=buf;
  str+=", Results: ";

  for (i =0 ; i < DATA_SIZE; i++) 
  {
    _gcvt((float)data[i],10, buf);
    str+=buf; str+=" ";
  }
  str+="\n";

  printf("%s", str.c_str());
};

////////////////////////////////////////////////////////////////////////////
int random(int num)
{
  return rand() % num;
};

////////////////////////////////////////////////////////////////////////////
MPI_Datatype randMPIType()
{
  switch (random(3))
  {
    case 0: return MPI_INT; break;
    case 1: return MPI_CHAR; break;
    case 2: return MPI_FLOAT; break;
  };

  return MPI_FLOAT;
};

////////////////////////////////////////////////////////////////////////////
// Prepare random communications map
void prepareRandomMap(int size)
{
  int i,j,k, iters;
  int p1, p2;
  Operation op;

  for (j =0; j <TOTAL_STEPS; j++)
  {
    // Make point-to-point random operations
    iters = (random(POINT_RAND) * size) / 2;
    for (i =0; i < iters; i++)
    {
      p1 = random(size);
      while ((p2 = random(size)) == p1) ;

      if (random(2) == 0) // Send & Recv
      {
        op.code = OP_SEND;
        op.delay = random(MAX_DELAY_TIME);
        op.destRank = p2;
        op.opType = randMPIType();
        map[p1].push_back(op);

        op.code = OP_RECV;
        op.delay = random(MAX_DELAY_TIME);
        op.destRank = p1;
        map[p2].push_back(op);
      }
      else // Sendrecv
      {
        op.code = OP_SENDRECV;
        op.delay = random(MAX_DELAY_TIME);
        op.destRank = p2;
        op.opType = randMPIType();
        map[p1].push_back(op);

        op.code = OP_SENDRECV;
        op.delay = random(MAX_DELAY_TIME);
        op.destRank = p1;
        map[p2].push_back(op);
      };
    };

    // Make broadcast random operations
    iters = random(GROUP_RAND);
    for (i =0; i < iters; i++)
    {
      switch (random(4))
      {
        case 0: op.code = OP_BCAST; break;
        case 1: op.code = OP_REDUCE; break;
        case 2: op.code = OP_ALLREDUCE; break;
        case 3: op.code = OP_ALLGATHER; break;
      };

      op.destRank = random(size);
      op.opType = randMPIType();

      for (k=0; k < size; k++) 
      {
        op.delay = random(MAX_DELAY_TIME);
        map[k].push_back(op);
      }
    };
  }

};


////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
  int rank, size;
  int i,j, k;
  int Tag = 10;
  MPI_Status status;
  Operation op;
  
  FILE * outFile = stdout;

  srand( (unsigned)time( NULL ) );

  if (MPI_Init(&argc, &argv) != MPI_SUCCESS)
  {
    MPI_Finalize();
    return 1;
  };

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
//size = 3;
//rank =1;

  if (rank == 0)
  {
    printf("MPI Test v1.0 started. \n Syntax GridTest [use]. \n 'Use' -  use already generated map, don't generate a new one. \n");
    if (argc > 1) if (strcmpi(argv[1],"use") == 0) GenerateNewMap = false;
  };

  if ((size > MAX_PROCESSES) || (size <2))
  {
    MPI_Finalize();
    printf("Illegal number of processes. MIN = 2, MAX = %d\n", MAX_PROCESSES);
    return 1;
  };

  if ((GenerateNewMap) && (rank == 0))  // Prepare random communications map
  {
    prepareRandomMap(size);

    // Print all to the file
    FILE *fp = fopen("newTest.map","wt");
    if (fp == NULL) 
    {
      MPI_Finalize();
      printf("Error opening output file newTest.map \n");
      return 1;
    };

    k = -1;    // find the longest queue
    for (i =0 ;i < size; i++)
    {
      if ((int)map[i].size() > k )  k = map[i].size();
    }

    for (i =0 ;i < k; i++)
    {
      for (j =0; j <size; j++)
      { 
        if (map[j].size() <= i) fprintf (fp, "%d\n%d\n%d\n%d\n", OP_EMPTY, 0, 0, 0);
        else fprintf (fp, "%d\n%d\n%d\n%d\n", map[j][i].code, map[j][i].destRank, map[j][i].opType, map[j][i].delay);
      }
    };
    fclose(fp);
    MPI_Barrier(MPI_COMM_WORLD);
  }

  if ((GenerateNewMap) && (rank == 0)) ;
  else  // Load the communication map
  {
    char str[100];

    MPI_Barrier(MPI_COMM_WORLD);

    FILE *fp = fopen("newTest.map","rt");
    if (fp == NULL) 
    {
      MPI_Finalize();
      printf("Error opening input file\n");
      return 1;
    };
    while (true)
    {
      for (j =0; j <size; j++)
      { 
        if (fgets(str, 100, fp) == NULL) break;
        op.code = atoi(str);
        if (fgets(str, 100, fp) == NULL) break;
        op.destRank = atoi(str);
        if (fgets(str, 100, fp) == NULL) break;
        op.opType = atoi(str);
        if (fgets(str, 100, fp) == NULL) break;
        op.delay = atoi(str);
        map[j].push_back(op);
      }

      if (j != size) break;
    };
    fclose(fp);
  };

  //char nm[20];
  //itoa(rank,nm,10);
  //outFile = fopen(nm,"wt");

  // Do communications
  for (i =0 ; i < DATA_SIZE; i++) data[i] = (float)((rank+1) * pow(10 ,i+1));
  printf("Task rank %d started\n", rank);

//  if (rank == 1)
  //for (i =0; i < map[0].size(); i++)
 // for (i =0; i < 5; i++)
//   printf ("%d %d %d %d\n", map[rank][i].code, map[rank][i].destRank, map[rank][i].opType, map[rank][i].delay);
 //     MPI_Finalize();
//return 1;

  for (i =0; i < map[0].size(); i++)
  {
    //printf("Rank = %d, op code = %d\n", rank, map[rank][i].code);
    switch (map[rank][i].code)
    {
      case OP_EMPTY: break;
      case OP_SEND: fprintf(outFile, "Rank = %d, MPI_Send to %d \n", rank, map[rank][i].destRank);fflush(outFile);
                    MPI_Send(&data,DATA_SIZE,MPI_FLOAT,map[rank][i].destRank,Tag,MPI_COMM_WORLD);
                    break;
      case OP_RECV: fprintf(outFile, "Rank = %d, MPI_Recv from %d \n", rank, map[rank][i].destRank);fflush(outFile);
                    MPI_Recv(&data,DATA_SIZE,MPI_FLOAT,map[rank][i].destRank,Tag,MPI_COMM_WORLD,&status);
                    break;
      case OP_SENDRECV: fprintf(outFile, "Rank = %d, MPI_Sendrecv from %d \n", rank, map[rank][i].destRank);fflush(outFile);
                        MPI_Sendrecv(&data, DATA_SIZE,MPI_FLOAT,map[rank][i].destRank,Tag,
                                     &data1, DATA_SIZE,MPI_FLOAT,map[rank][i].destRank,Tag, MPI_COMM_WORLD, &status);
                        break;
      case OP_REDUCE: fprintf(outFile, "Rank = %d, MPI_Reduce \n", rank);fflush(outFile);
                      MPI_Reduce(data, data1, DATA_SIZE, MPI_FLOAT, MPI_SUM, map[rank][i].destRank,  MPI_COMM_WORLD);
                      if (rank == map[rank][i].destRank) 
                        for (j=0; j < DATA_SIZE; j++) {data[j] = data1[j]; data[j] /=(size-1);}
                      break;
      case OP_ALLREDUCE: fprintf(outFile, "Rank = %d, MPI_Allreduce \n", rank);fflush(outFile);
                         MPI_Allreduce(data, data1, DATA_SIZE, MPI_FLOAT, MPI_SUM,  MPI_COMM_WORLD);
                         for (j=0; j < DATA_SIZE; j++) {data[j] = data1[j]; data[j] /=(size-1);}
                         break;
      case OP_BCAST: fprintf(outFile, "Rank = %d, MPI_bcast \n", rank);fflush(outFile);
                     MPI_Bcast (data, DATA_SIZE, MPI_FLOAT, map[rank][i].destRank, MPI_COMM_WORLD);
                     break;
      case OP_ALLGATHER: fprintf(outFile, "Rank = %d, MPI_Allgather \n", rank);fflush(outFile);
                         MPI_Allgather(data, DATA_SIZE,MPI_FLOAT, data2, DATA_SIZE, MPI_FLOAT, MPI_COMM_WORLD);
                         break;
      default: printf("Error in Op CODE !! \n"); break;
    };
    fprintf(outFile, "Rank %d, step %d done -----------------------\n", rank, i);fflush(outFile);

    Sleep(map[rank][i].delay);
  };

  MPI_Barrier(MPI_COMM_WORLD);

  writeResults(rank);

  MPI_Finalize();
  return 0;
}
