
#include <stdio.h>

#include "grid_mpi.h"
#include "types.h"
#include "../Shared/multi_os.h"

int gridType, gridType1;
MPI_Datatype MPI_type, MPI_type1;
int memSize;
int cOp;
MPI_Op mpiOp;

#ifdef _WIN32
  #define FMPG_Init1 MPG_INIT
  #define FMPG_Comm_size1 MPG_COMM_SIZE
  #define FMPG_Comm_rank1 MPG_COMM_RANK
  #define FMPG_Send1 MPG_SEND
  #define FMPG_Recv1 MPG_RECV
  #define FMPG_Finalize1 MPG_FINALIZE
  #define FMPG_Sendrecv1 MPG_SENDRECV
  #define FMPG_Bcast1 MPG_BCAST
  #define FMPG_Reduce1 MPG_REDUCE
  #define FMPG_Allreduce1 MPG_ALLREDUCE
  #define FMPG_Allgather1 MPG_ALLGATHER
  #define FMPG_Barrier1 MPG_BARRIER

  #define FMPG_Pack1 MPG_PACK
  #define FMPG_Unpack1 MPG_UNPACK
  
  #define FConvertG_F CONVGF
  #define FConvertF_G CONVFG
  
  #define FConvertMPI_OP CONVOP
  
#else
  // Diffirent Fortran compilers add 0-1-2 '_' symbols to functions names . We should support all these cases
  #define FMPG_Init1 mpg_init_
  #define FMPG_Comm_size1 mpg_comm_size_
  #define FMPG_Comm_rank1 mpg_comm_rank_
  #define FMPG_Send1 mpg_send_
  #define FMPG_Recv1 mpg_recv_
  #define FMPG_Finalize1 mpg_finalize_
  #define FMPG_Sendrecv1 mpg_sendrecv_
  #define FMPG_Bcast1 mpg_bcast_
  #define FMPG_Reduce1 mpg_reduce_
  #define FMPG_Allreduce1 mpg_allreduce_
  #define FMPG_Allgather1 mpg_allgather_
  #define FMPG_Barrier1 mpg_barrier_
  #define FMPG_Pack1 mpg_pack_
  #define FMPG_Unpack1 mpg_unpack_

  #define FMPG_Init2 mpg_init__
  #define FMPG_Comm_size2 mpg_comm_size__
  #define FMPG_Comm_rank2 mpg_comm_rank__
  #define FMPG_Send2 mpg_send__
  #define FMPG_Recv2 mpg_recv__
  #define FMPG_Finalize2 mpg_finalize__
  #define FMPG_Sendrecv2 mpg_sendrecv__
  #define FMPG_Bcast2 mpg_bcast__
  #define FMPG_Reduce2 mpg_reduce__
  #define FMPG_Allreduce2 mpg_allreduce__
  #define FMPG_Allgather2 mpg_allgather__
  #define FMPG_Barrier2 mpg_barrier__
  #define FMPG_Pack2 mpg_pack__
  #define FMPG_Unpack2 mpg_unpack__

  #define FMPG_Init3 mpg_init
  #define FMPG_Comm_size3 mpg_comm_size
  #define FMPG_Comm_rank3 mpg_comm_rank
  #define FMPG_Send3 mpg_send
  #define FMPG_Recv3 mpg_recv
  #define FMPG_Finalize3 mpg_finalize
  #define FMPG_Sendrecv3 mpg_sendrecv
  #define FMPG_Bcast3 mpg_bcast
  #define FMPG_Reduce3 mpg_reduce
  #define FMPG_Allreduce3 mpg_allreduce
  #define FMPG_Allgather3 mpg_allgather
  #define FMPG_Barrier3 mpg_barrier
  #define FMPG_Pack3 mpg_pack
  #define FMPG_Unpack3 mpg_unpack
  
  #define FConvertG_F convgf_
  #define FConvertF_G convfg_

  #define FConvertMPI_OP convop_
  
#endif

extern "C"
{
  void _CALL FConvertG_F(int &dForType, int &dGridType);
  void _CALL FConvertF_G(int &dForType, int &dGridType);
  void _CALL FConvertMPI_OP(int &forOp, int &cOp);
};


////////////////////////////////////////////////////////////////////////////
// implementations for Fortran
  void FMPG_Finalize(int *ierr)
  {
    MPG_Finalize();
  }

  void FMPG_Comm_size(MPI_Comm *comm, int *size, int *ierr)
  {
    *ierr = MPG_Comm_size(*comm, size);
  }

  void FMPG_Comm_rank(MPI_Comm *comm, int *rank, int *ierr)
  {
    *ierr = MPG_Comm_rank(*comm, rank);
  }

  void FMPG_Recv(void *buf, int *count, int *datatype, int *source, int *tag, MPI_Comm *comm, MPI_Status *status, int *ierr)
  {
    FConvertF_G(*datatype, gridType);
    if (gridType == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *datatype);
    convertGrid_MPI(gridType, MPI_type, memSize);

    *ierr = MPG_Recv(buf, *count, MPI_type, *source, *tag, *comm, status);
  }

  void FMPG_Send(void *buf, int *count, int *datatype, int *dest, int *tag, MPI_Comm *comm, int *ierr)
  {
    FConvertF_G(*datatype, gridType);
    if (gridType == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *datatype);
    convertGrid_MPI(gridType, MPI_type, memSize);

    *ierr = MPG_Send(buf, *count, MPI_type, *dest, *tag, *comm);
  }

  void FMPG_Sendrecv(void *sendbuf, int *sendcount, int *sendtype, int *dest, int *sendtag,void *recvbuf,
                     int *recvcount, int *recvtype, int *source, int *recvtag, MPI_Comm *comm,
                     MPI_Status *status, int *ierr)
  {
    FConvertF_G(*sendtype, gridType);
    if (gridType == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *sendtype);
    convertGrid_MPI(gridType, MPI_type, memSize);

    FConvertF_G(*recvtype, gridType1);
    if (gridType1 == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *recvtype);
    convertGrid_MPI(gridType1, MPI_type1, memSize);

    *ierr = MPG_Sendrecv(sendbuf, *sendcount, MPI_type, *dest, *sendtag, recvbuf, *recvcount,
                         MPI_type1, *source, *recvtag, *comm, status);
  }

  void FMPG_Barrier(MPI_Comm *comm, int *ierr)
  {
    *ierr = MPG_Barrier(*comm);
  }

  void FMPG_Reduce(void *sbuf, void *rbuf, int *count, int *datatype, int *op, int *root, MPI_Comm *comm, int *ierr)
  {
    FConvertF_G(*datatype, gridType);
    if (gridType == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *datatype);

    FConvertMPI_OP(*op, cOp);
    if (cOp == -1) printf("ERROR. Unsupported MPI_OP type (Fortran)in MPI_Reduce\n");
    convertMPI_OP_C(cOp, mpiOp);
    
    convertGrid_MPI(gridType, MPI_type, memSize);

    *ierr = MPG_Reduce(sbuf, rbuf, *count, MPI_type, mpiOp, *root, *comm);
  }

  void FMPG_Allreduce(void *sbuf, void *rbuf, int *count, int *datatype, int *op, MPI_Comm *comm, int *ierr)
  {
    FConvertF_G(*datatype, gridType);
    if (gridType == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *datatype);
    
    FConvertMPI_OP(*op, cOp);
    if (cOp == -1) printf("ERROR. Unsupported MPI_OP type (Fortran)in MPI_Allreduce\n");
    convertMPI_OP_C(cOp, mpiOp);
    
    convertGrid_MPI(gridType, MPI_type, memSize);

    *ierr = MPG_Allreduce(sbuf, rbuf, *count, MPI_type, mpiOp, *comm);
  }

  void FMPG_Allgather(void *send, int *sCount, int *sType, void *recv, int *rCount, int *rType, MPI_Comm *comm, int *ierr)
  {
    FConvertF_G(*sType, gridType);
    if (gridType == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *sType);
    convertGrid_MPI(gridType, MPI_type, memSize);

    *ierr = MPG_Allgather(send, *sCount, MPI_type, recv, *rCount, MPI_type, *comm);
  }

  void FMPG_Bcast(void *buf, int *count, int *datatype, int *root, MPI_Comm *comm, int *ierr)
  {
    FConvertF_G(*datatype, gridType);
    if (gridType == -1) printf("ERROR. Unsupported MPI type (Fortran), MPI_Datatype = %d\n", *datatype);
    convertGrid_MPI(gridType, MPI_type, memSize);

    *ierr = MPG_Bcast (buf, *count, MPI_type, *root, *comm);
  }
  void FMPG_Pack()
  {
    printf("ERROR. MPG_Pack: not implemented.\n");
  }
  void FMPG_Unpack()
  {
    printf("ERROR. MPG_Unpack: not implemented.\n");
  }


////////////////////////////////////////////////////////////////////////////
// definitions for Fortran
extern "C"
{
 // 1st case --
  void _CALL FMPG_Finalize1(int *ierr)
  { FMPG_Finalize(ierr); }

  void _CALL FMPG_Comm_size1(MPI_Comm *comm, int *size, int *ierr)
  { FMPG_Comm_size(comm, size, ierr); }

  void _CALL FMPG_Comm_rank1(MPI_Comm *comm, int *rank, int *ierr)
  { FMPG_Comm_rank(comm, rank, ierr); }

  void _CALL FMPG_Recv1(void *buf, int *count, int *datatype, int *source, int *tag, MPI_Comm *comm, MPI_Status *status, int *ierr)
  { FMPG_Recv(buf, count, datatype, source, tag, comm, status, ierr); }

  void _CALL FMPG_Send1(void *buf, int *count, int *datatype, int *dest, int *tag, MPI_Comm *comm, int *ierr)
  { FMPG_Send(buf, count, datatype, dest, tag, comm, ierr); }

  void _CALL FMPG_Sendrecv1(void *sendbuf, int *sendcount, int *sendtype, int *dest, int *sendtag,void *recvbuf,
                     int *recvcount, int *recvtype, int *source, int *recvtag, MPI_Comm *comm,
                     MPI_Status *status, int *ierr)
  { FMPG_Sendrecv(sendbuf, sendcount, sendtype, dest, sendtag,recvbuf, recvcount, recvtype,
                  source, recvtag, comm, status, ierr); }

  void _CALL FMPG_Barrier1(MPI_Comm *comm, int *ierr)
  { FMPG_Barrier(comm, ierr); }

  void _CALL FMPG_Reduce1(void *sbuf, void *rbuf, int *count, int *datatype, int *op, int *root, MPI_Comm *comm, int *ierr)
  { FMPG_Reduce(sbuf, rbuf, count, datatype, op, root, comm, ierr); }

  void _CALL FMPG_Allreduce1(void *sbuf, void *rbuf, int *count, int *datatype, int *op, MPI_Comm *comm, int *ierr)
  { FMPG_Allreduce(sbuf, rbuf, count, datatype, op, comm, ierr); }

  void _CALL FMPG_Allgather1(void *send, int *sCount, int *sType, void *recv, int *rCount, int *rType, MPI_Comm *comm, int *ierr)
  { FMPG_Allgather(send, sCount, sType, recv, rCount, rType, comm, ierr); }

  void _CALL FMPG_Bcast1(void *buf, int *count, int *datatype, int *root, MPI_Comm *comm, int *ierr)
  { FMPG_Bcast(buf, count, datatype, root, comm, ierr); }

  void _CALL FMPG_Pack1()
  { FMPG_Pack(); }
  void _CALL FMPG_Unpack1()
  { FMPG_Unpack(); }

 // 2nd case --
  void _CALL FMPG_Finalize2(int *ierr)
  { FMPG_Finalize(ierr); }

  void _CALL FMPG_Comm_size2(MPI_Comm *comm, int *size, int *ierr)
  { FMPG_Comm_size(comm, size, ierr); }

  void _CALL FMPG_Comm_rank2(MPI_Comm *comm, int *rank, int *ierr)
  { FMPG_Comm_rank(comm, rank, ierr); }

  void _CALL FMPG_Recv2(void *buf, int *count, int *datatype, int *source, int *tag, MPI_Comm *comm, MPI_Status *status, int *ierr)
  { FMPG_Recv(buf, count, datatype, source, tag, comm, status, ierr); }

  void _CALL FMPG_Send2(void *buf, int *count, int *datatype, int *dest, int *tag, MPI_Comm *comm, int *ierr)
  { FMPG_Send(buf, count, datatype, dest, tag, comm, ierr); }

  void _CALL FMPG_Sendrecv2(void *sendbuf, int *sendcount, int *sendtype, int *dest, int *sendtag,void *recvbuf,
                     int *recvcount, int *recvtype, int *source, int *recvtag, MPI_Comm *comm,
                     MPI_Status *status, int *ierr)
  { FMPG_Sendrecv(sendbuf, sendcount, sendtype, dest, sendtag,recvbuf, recvcount, recvtype,
                  source, recvtag, comm, status, ierr); }

  void _CALL FMPG_Barrier2(MPI_Comm *comm, int *ierr)
  { FMPG_Barrier(comm, ierr); }

  void _CALL FMPG_Reduce2(void *sbuf, void *rbuf, int *count, int *datatype, int *op, int *root, MPI_Comm *comm, int *ierr)
  { FMPG_Reduce(sbuf, rbuf, count, datatype, op, root, comm, ierr); }

  void _CALL FMPG_Allreduce2(void *sbuf, void *rbuf, int *count, int *datatype, int *op, MPI_Comm *comm, int *ierr)
  { FMPG_Allreduce(sbuf, rbuf, count, datatype, op, comm, ierr); }

  void _CALL FMPG_Allgather2(void *send, int *sCount, int *sType, void *recv, int *rCount, int *rType, MPI_Comm *comm, int *ierr)
  { FMPG_Allgather(send, sCount, sType, recv, rCount, rType, comm, ierr); }

  void _CALL FMPG_Bcast2(void *buf, int *count, int *datatype, int *root, MPI_Comm *comm, int *ierr)
  { FMPG_Bcast(buf, count, datatype, root, comm, ierr); }

  void _CALL FMPG_Pack2()
  { FMPG_Pack(); }
  void _CALL FMPG_Unpack2()
  { FMPG_Unpack(); }

 // 3d case --
  void _CALL FMPG_Finalize3(int *ierr)
  { FMPG_Finalize(ierr); }

  void _CALL FMPG_Comm_size3(MPI_Comm *comm, int *size, int *ierr)
  { FMPG_Comm_size(comm, size, ierr); }

  void _CALL FMPG_Comm_rank3(MPI_Comm *comm, int *rank, int *ierr)
  { FMPG_Comm_rank(comm, rank, ierr); }

  void _CALL FMPG_Recv3(void *buf, int *count, int *datatype, int *source, int *tag, MPI_Comm *comm, MPI_Status *status, int *ierr)
  { FMPG_Recv(buf, count, datatype, source, tag, comm, status, ierr); }

  void _CALL FMPG_Send3(void *buf, int *count, int *datatype, int *dest, int *tag, MPI_Comm *comm, int *ierr)
  { FMPG_Send(buf, count, datatype, dest, tag, comm, ierr); }

  void _CALL FMPG_Sendrecv3(void *sendbuf, int *sendcount, int *sendtype, int *dest, int *sendtag,void *recvbuf,
                     int *recvcount, int *recvtype, int *source, int *recvtag, MPI_Comm *comm,
                     MPI_Status *status, int *ierr)
  { FMPG_Sendrecv(sendbuf, sendcount, sendtype, dest, sendtag,recvbuf, recvcount, recvtype,
                  source, recvtag, comm, status, ierr); }

  void _CALL FMPG_Barrier3(MPI_Comm *comm, int *ierr)
  { FMPG_Barrier(comm, ierr); }

  void _CALL FMPG_Reduce3(void *sbuf, void *rbuf, int *count, int *datatype, int *op, int *root, MPI_Comm *comm, int *ierr)
  { FMPG_Reduce(sbuf, rbuf, count, datatype, op, root, comm, ierr); }

  void _CALL FMPG_Allreduce3(void *sbuf, void *rbuf, int *count, int *datatype, int *op, MPI_Comm *comm, int *ierr)
  { FMPG_Allreduce(sbuf, rbuf, count, datatype, op, comm, ierr); }

  void _CALL FMPG_Allgather3(void *send, int *sCount, int *sType, void *recv, int *rCount, int *rType, MPI_Comm *comm, int *ierr)
  { FMPG_Allgather(send, sCount, sType, recv, rCount, rType, comm, ierr); }

  void _CALL FMPG_Bcast3(void *buf, int *count, int *datatype, int *root, MPI_Comm *comm, int *ierr)
  { FMPG_Bcast(buf, count, datatype, root, comm, ierr); }

  void _CALL FMPG_Pack3()
  { FMPG_Pack(); }
  void _CALL FMPG_Unpack3()
  { FMPG_Unpack(); }
}

