#include <mpi.h>
#include <stdio.h>

#include "types.h"
#include "../Shared/multi_os.h"

///////////////////////////////////////////////////////////////////////////////
int convertMPI_Grid(MPI_Datatype MPI_type, int &gridType, int &memSize)
{
  // check C types
  if (MPI_type == MPI_CHAR)
  {
    memSize =sizeof(signed char);
    gridType = TYPE_CHAR;
    return 1;
  }
  if (MPI_type == MPI_SHORT)
  {
    memSize =sizeof(signed short int);
    gridType = TYPE_SHORT;
    return 1;
  }
  if (MPI_type == MPI_INT)
  {
    memSize =sizeof(signed int);
    gridType = TYPE_INT;
    return 1;
  }    
  if (MPI_type == MPI_LONG)
  {
    memSize =sizeof(signed long int);
    gridType = TYPE_LONG;
    return 1;
  }    
  if (MPI_type == MPI_UNSIGNED_CHAR)
  {
    memSize =sizeof(unsigned char);
    gridType = TYPE_UCHAR;
    return 1;
  }    
  if (MPI_type == MPI_UNSIGNED_SHORT)
  {
    memSize =sizeof(unsigned short int);
    gridType = TYPE_USHORT;
    return 1;
  }    
  if (MPI_type == MPI_UNSIGNED)
  {
    memSize =sizeof(unsigned int);
    gridType = TYPE_UNSIGNED;
    return 1;
  }    
  if (MPI_type == MPI_UNSIGNED_LONG)
  {
    memSize =sizeof(unsigned long int);
    gridType = TYPE_ULONG;
    return 1;
  }    
  if (MPI_type == MPI_FLOAT)
  {
    memSize =sizeof(float);
    gridType = TYPE_FLOAT;
    return 1;
  }    
  if (MPI_type == MPI_DOUBLE)
  {
    memSize =sizeof(double);
    gridType = TYPE_DOUBLE;
    return 1;
  }    
  if (MPI_type == MPI_LONG_DOUBLE)
  {
    memSize =sizeof(long double);
    gridType = TYPE_LDOUBLE;
    return 1;
  }    
  if (MPI_type == MPI_BYTE)
  {
    printf("ERROR. Unsupported MPI type - MPI_BYTE in a packet.\n");
    return 0;
  }
  if (MPI_type == MPI_PACKED)
  {
    printf("ERROR. Unsupported MPI type - MPI_PACKED in a packet.\n");
    return 0;
  }
  
  printf("ERROR. Unknown MPI type %d in a packet.\n",MPI_type);
  return 0;
  
}

///////////////////////////////////////////////////////////////////////////////
int convertGrid_MPI(int gridType, MPI_Datatype &MPI_type, int &memSize)
{
  // check C types
  switch (gridType)
  {
    case TYPE_CHAR        :memSize =sizeof(signed char);
                           MPI_type = MPI_CHAR;
                           break;
    case TYPE_SHORT       :memSize =sizeof(signed short int);
                           MPI_type = MPI_SHORT;
                           break;
    case TYPE_INT         :memSize =sizeof(signed int);
                           MPI_type = MPI_INT;
                           break;
    case TYPE_LONG        :memSize =sizeof(signed long int);
                           MPI_type = MPI_LONG;
                           break;
    case TYPE_UCHAR       :memSize =sizeof(unsigned char);
                           MPI_type = MPI_UNSIGNED_CHAR;
                           break;
    case TYPE_USHORT      :memSize =sizeof(unsigned short int);
                           MPI_type = MPI_UNSIGNED_SHORT;
                           break;
    case TYPE_UNSIGNED    :memSize =sizeof(unsigned int);
                           MPI_type = MPI_UNSIGNED;
                           break;
    case TYPE_ULONG       :memSize =sizeof(unsigned long int);
                           MPI_type = MPI_UNSIGNED_LONG;
                           break;
    case TYPE_FLOAT       :memSize =sizeof(float);
                           MPI_type = MPI_FLOAT;
                           break;
    case TYPE_DOUBLE      :memSize =sizeof(double);
                           MPI_type = MPI_DOUBLE;
                           break;
    case TYPE_LDOUBLE     :memSize =sizeof(long double);
                           MPI_type = MPI_LONG_DOUBLE;
                           break;

    default               :printf("ERROR. Unknown GRID type %d in a packet.\n",gridType);
                           return 0;
  }

  return 1;
}

int convertMPI_OP_C(int code, MPI_Op &op)
{
  switch (code)
  {
    case MPI_OP_MAX: op = MPI_MAX; break;
    case MPI_OP_MIN: op = MPI_MIN; break;
    case MPI_OP_SUM: op = MPI_SUM; break;
    case MPI_OP_PROD: op = MPI_PROD; break;
    case MPI_OP_LAND: op = MPI_LAND; break;
    case MPI_OP_BAND: op = MPI_BAND; break;
    case MPI_OP_LOR: op = MPI_LOR; break;
    case MPI_OP_BOR: op = MPI_BOR; break;
    case MPI_OP_LXOR: op = MPI_LXOR; break;
    case MPI_OP_BXOR: op = MPI_BXOR; break;
    case MPI_OP_MAXLOC: op = MPI_MAXLOC; break;
    case MPI_OP_MINLOC: op = MPI_MINLOC; break;
    default: return 0;
  };

  return 1;
};
