
// Main server module

int DEBUG_LEVEL  = 0 ; // Define this variable to see debug messages

#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "../Shared/multi_os.h"

#ifdef _WIN32
  #include <direct.h>
  #include <windows.h>
#else
#endif

#include "server.h"
#include "../Shared/taskctrl.h"
#include "../Shared/net.h"

Server server;

void writeHelp()
{
  printf("\n");
  printf("NumGRID jobmanager syntax :\n");
  printf("jobmanager [confFile] [servIndex]\n");
  printf("[confFile] - Configuration file with user's task description to load (default  - task.ng)\n");
  printf("[servIndex] - A position of this cluster in the conf file (int :0,1,2,.., default :auto)\n");
};

///////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
  std::string runScriptName = DEFAULT_JOB_FILENAME;

  printf("NumGRID jobmanager. Version 1.0. (c) SSD, 2007\n");

  if (getNgPath() == "") return 1;
  if ( !initNetwork()) return 1;

  server.init();

  // parse command line args
  // [1st arg] = 'run script' path & name
  // [2nd arg] = this server's position in a server's list

  if (argc >1)
  {
    runScriptName = argv[1];
  }

  printf("Loading new task. Task conf file = '%s'...", runScriptName.c_str());

  if (server.project.LoadInfo((char *)runScriptName.c_str()))
    printf("OK. \n");
  else
  {
    writeHelp();
    return 1;
  }

  server.ngPath = getNgPath();

  server.establishNumeration();
//  server.showJobParams();

  if (argc >2) 
  {
    int myInd = atoi(argv[2]);
    server.myIndex = myInd;
  }
  else
    if (!server.findThis()) 
    {
      writeHelp();
      return 1;
    }
    
  if ((server.myIndex < 0) || (server.myIndex >= server.project.serverList.size()))
  {
    printf("ERROR. Wrong index( %d) for jobmanager. Should be in a range [0..maxServers]\n",server.myIndex);
    return 1;
  };

  printf("Starting gateway service at host: %s(%s):%d, supported ranks :%d-%d...",
         server.project.serverList[server.myIndex].serverName.c_str(), server.project.serverList[server.myIndex].internalHostName.c_str(),
         server.project.serverList[server.myIndex].gatewayPort, server.project.serverList[server.myIndex].minRank,
         server.project.serverList[server.myIndex].maxRank);
	 
  if ((server.servSoc = makeServerSocket(server.project.serverList[server.myIndex].gatewayPort)) < 0) 
  {
    printf("ERROR. Can't create socket on port %d. (makeServerSocket)\n",server.project.serverList[server.myIndex].gatewayPort);
    return 1;
  }

  printf("OK\n");
  
  char dir[255];
  getcwd(dir,255);
  printf("Current directory is %s\n", dir);
  // Start workers
  if (server.project.serverList[server.myIndex].workDir.compare("") != 0)
     if (chdir(server.project.serverList[server.myIndex].workDir.c_str()) == -1)
      printf("WARNING. Trying to change directory to %s ,path could not be found. check %s script\n",
             server.project.serverList[server.myIndex].workDir.c_str(), runScriptName.c_str());

  server.taskCtrl.init(server.project, server.myIndex, getNgPath());

  if (!server.taskCtrl.loadServerInfo(SERVER_CFG_FILE)) return 1;
  
  if (server.project.serverList[server.myIndex].needRunTask)
  {
    if (!server.taskCtrl.run()) return 1;
  };

  // Connect all gateways
  printf("Trying to connect with all other gateway servers....");
  if ( !server.connectAllServers() ) return 1;
  printf("OK\n");

  server.run();
  
  closeNetwork();
  return 0;
};
