
// main server module

#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "../Shared/multi_os.h"

#ifdef _WIN32
  #include <direct.h>
  #include <time.h>
#else
  #include <sched.h>
  #include <signal.h>
  #include <unistd.h>
#endif

#include "server.h"

///////////////////////////////////////////////////////////////////////////////
// Write thread. Sends packets to their receivers
void Server::writePackets()
{
  int i;

  for (i =0; i < links.size(); i++)
  {
    if (links[i].packets.size() !=0)
    {
//      showPacketInfo(serv->links[i].packets.front());
        if (deliverPacket(links[i].packets.front(), links[i].service))
        {
          free(links[i].packets.front().data);
//          printf("packet sent to rank %d\n", links[i].packets.front().header.recvAddr);
//          fflush(stdout);
          links[i].packets.erase(links[i].packets.begin());
        }
      }
  };
};

///////////////////////////////////////////////////////////////////////////////
Server::Server()
{
}

///////////////////////////////////////////////////////////////////////////////
// Get current directory
char *Server::getWorkDir()
{
  static char buf[MAX_DIRECTORY_LENGTH];
  return getcwd(buf, MAX_DIRECTORY_LENGTH);
};

///////////////////////////////////////////////////////////////////////////////
// Print job parameters
void Server::showJobParams()
{
  int i,j;

  printf("\n------------------------------\n");
  for (i =0; i< project.serverList.size(); i++)
  {
    printf("blockName = %s \n",project.serverList[i].blockName.c_str());
    printf("serverName = %s \n",project.serverList[i].serverName.c_str());
    printf("internalHostName = %s \n",project.serverList[i].internalHostName.c_str());
    printf("serverPort = %d \n",project.serverList[i].serverPort);
    printf("gatewayPort = %d \n",project.serverList[i].gatewayPort);

    printf("username = %s \n",project.serverList[i].username.c_str());
    printf("password = %s \n",project.serverList[i].password.c_str());

    printf("nProcesses = %d \n",project.serverList[i].nProcesses);

    printf("needRunTask = %d \n",project.serverList[i].needRunTask);
    printf("executable = %s \n",project.serverList[i].executable.c_str());

    printf("makefile = %s \n",project.serverList[i].makefile.c_str());

    printf("localSourceDir = %s \n",project.serverList[i].sourceDir.c_str());
    printf("localDataDir = %s \n",project.serverList[i].inputDir.c_str());
    printf("localResultDir = %s \n",project.serverList[i].outputDir.c_str());
    printf("workDir = %s \n",project.serverList[i].workDir.c_str());

    printf("minRank = %d \n",project.serverList[i].minRank);
    printf("maxRank = %d \n",project.serverList[i].maxRank);

    printf("sourceFiles = ");
    for (j= 0; j < project.serverList[i].sourceFiles.size(); j++)
      printf("%s ",project.serverList[i].sourceFiles[j].c_str());

    printf("\ninputFiles = ");
    for (j= 0; j < project.serverList[i].inputFiles.size(); j++)
      printf("%s ",project.serverList[i].inputFiles[j].c_str());

    printf("\noutputFiles = ");
    for (j= 0; j < project.serverList[i].outputFiles.size(); j++)
      printf("%s ",project.serverList[i].outputFiles[j].c_str());
    printf("\n------------------------------\n");
  }

}

///////////////////////////////////////////////////////////////////////////////
// Establish new numeration in server's list
void Server::establishNumeration()
{
  project.serverList[0].minRank = 0;
  project.serverList[0].maxRank = project.serverList[0].nProcesses - 1;

  for (int i =1; i< project.serverList.size(); i ++)
  {
    project.serverList[i].minRank = project.serverList[i-1].maxRank + 1;
    project.serverList[i].maxRank = project.serverList[i].minRank + project.serverList[i].nProcesses - 1;
  }
};

///////////////////////////////////////////////////////////////////////////////
// Locate this server in server's list
// if successful returns 1. in case of error returns 0
int Server::findThis()
{
  char serverName[MAX_HOSTNAME_LENGTH];

  if (gethostname(serverName,MAX_HOSTNAME_LENGTH)  != 0)
  {
    printf("ERROR. findThis(). Can't resolve the host name %s. \n", serverName);
    return 0;
  }

  myIndex = -1;
  for (int i =0; i< project.serverList.size(); i ++)
  {
    if (project.serverList[i].internalHostName.compare(serverName) == 0)
      myIndex = i;
  }

  if (myIndex == -1) 
  {
    printf("ERROR. Can't locate this host(%s) in the servers list in the config file. Check ServerName & InternalHostName definitions\n", serverName);
    return 0;
  }

  return 1;
}

///////////////////////////////////////////////////////////////////////////////
// Initialize server.
// if successful returns 1. in case of warning returns 0
int Server::init()
{
  int i,j;

  myIndex = 0;
  serverPort = DEFAULT_SERVER_PORT;        

  links.clear();

  // connect all servers (FULL GPAGH topology by default)
  for (i =0; i < MAX_SERVERS; i++)
    for (j =0; j < MAX_SERVERS; j++)
      sTopologyInfo[i][j] = CONN;

  project.serverList.clear();

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
int Server::deliverP2Ppacket(Packet &pack)
{
  int i;

  i = findSendLink(pack.header.recvAddr);
  if (i ==-1) return 0;

  //printf("connected receiver found.sending to soc %d....\n", links[i].soc);
  if (writePacket(links[i].soc, &pack) ==-1) // transfer packet to receiver
    return 0;
  else return 1;
};

///////////////////////////////////////////////////////////////////////////////
// Deliver packet to receiver (receiver address = pack.header.recvAddr)
// myNum - index of current thread
// if successful returns 1. in case of error returns 0
int Server::deliverPacket(Packet &pack, int sendService)
{
  int i;

  switch (pack.header.mesType)
  {

    // process broadcasts
    case MES_BROADCAST:
                   if (sendService == SERVICE_CLIENT) // send to all servers
                   {
                     for (i =0; i< links.size(); i++)
                     {
                       if ((links[i].service == SERVICE_SERVER) && (links[i].state == STATE_OPENED))
                       {
                        pack.header.recvAddr = links[i].minRank;
                        if (!deliverP2Ppacket(pack)) return 0;
                       }
                     }
                   }
                   else // send to to the client (local root)
                   {
                     if (!deliverP2Ppacket(pack)) return 0;
                   }

                   return 1;
                   break;

    case MES_P2P  :return deliverP2Ppacket(pack);
                   break;

    default       :printf("ERROR. deliverPacket(). Error in message type.\n");
                   showPacketInfo(pack);
                   terminateServer(SIG_TERMINATE);
  };

  printf("ERROR. deliverPacket(). Internal error. Contact with the developers\n");
  terminateServer(SIG_TERMINATE);  

  return 0;
};

///////////////////////////////////////////////////////////////////////////////
// Get information about new remote link (server)
// myNum - index of current thread
// links[myNum].soc contains already connectes socket to a new server 
// if successful returns 1. in case of error returns 0
int Server::getLinkInfo(LinkInfo *lnk)
{
  struct Packet pack;
  struct ConnectInfo conInfo;

  if (lnk == NULL) return 0;

  // Get info about new link
  pack.data = NULL; 
  if (readPacketAlloc(lnk->soc,&pack) == -1)
  {
    printf("ERROR. readPacketAlloc().\n");
    return 0;
  };

  // -------- check data
  if (pack.header.size != sizeof(struct ConnectInfo))
  {
    printf("ERROR. Unknown protocol.\n");
    return 0;
  }

  conInfo = *((struct ConnectInfo *)pack.data);
  if (conInfo.id != PROGRAM_ID)
  {
    printf("ERROR. Unknown indentifier. Program ID expected. \n");
    return 0;
  }

  if (conInfo.ngVersion != NG_VERSION)
  {
    printf("ERROR. The client uses an old version of NumGRID library. Please recompile you application program with the current NumGRID library, and check versions of all NumGRID gateways\n");
    return 0;
  }

  lnk->service = conInfo.service;
  lnk->minRank = conInfo.minRank;
  lnk->maxRank = conInfo.maxRank;

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// close link
void Server::closeLink(int index)
{
  closeSocket(links[index].soc); 
  printf("(Server link, socket = %d closed)\n",links[index].soc);

  links.erase(links.begin() + index);
};

///////////////////////////////////////////////////////////////////////////////
// Terminates server
void Server::terminateServer(int sig)
{
  int i;

  broadcastSignal(sig);
//  notifyAllServers(sig);

  for (i = links.size() -1 ; i> -1; i--)
    closeLink(i);

  links.clear();
  closeSocket(servSoc);

  closeNetwork();
  printf("Server terminated!.\n");
  exit(1);
};

///////////////////////////////////////////////////////////////////////////////
// Send message to all connected servers with pack.header.signal = sig
void Server::notifyAllServers(int sig)
{
  int i;
  Packet pack;

  for (i =0; i< links.size(); i++)
  {
    if ((links[i].service == SERVICE_SERVER) && (links[i].state == STATE_OPENED))
    {
      pack.header.signal = sig;
      pack.header.size =0;
      if (writePacket(links[i].soc, &pack) ==-1) 
        printf("Notification failed. notifyAllServers().\n");
    }
  }
};

///////////////////////////////////////////////////////////////////////////////
void Server::processSignal(int sig, int index)
{
  static Packet pack;
  static int workerTerminated = 0, serverTerminated = 0;
  static int linksLeft;

  //printf("------- %d ----\n",sig);

  switch (sig)
  {
    case SIG_NONE           :return;          
    case SIG_CLIENT_FINISHED:printf("Client rank = %d, socket = %d finished. closing this link....OK\n",
                                      links[index].minRank, links[index].soc);
                            links[index].state = STATE_NEEDCLOSE;
                            workerTerminated++;
                            break;

    case SIG_SERVER_FINISHED:serverTerminated++;
                             links[index].state = STATE_NEEDCLOSE;
                             break;

    case SIG_TERMINATE      :terminateServer(SIG_TERMINATE); break;
    case SIG_SYNHRONIZE     :printf("ERROR. Unsupported signal ID_SYNHRONIZE\n");
                             break;

    default                 :printf("ERROR. Unknown signal %d recived\n", sig);
                             break;
  };

  // check  if server shutdown needed
 linksLeft = project.serverList[myIndex].nProcesses+
             project.serverList.size() - myIndex -1 ;
 if (workerTerminated ==  project.serverList[myIndex].nProcesses )
 {
   if (serverTerminated == myIndex )
   {
     printf("Last link processed\n");
     // wait while all messages are sent
     while (!isMesQueueEmpty()); 

     terminateServer(SIG_SERVER_FINISHED);
   }
 };
};

///////////////////////////////////////////////////////////////////////////////
// Connect all servers at start-up
// if successful returns 1. in case of error returns 0
int Server::connectAllServers()
{
  int i;
  int newSoc;
  LinkInfo *lnk = new LinkInfo;

  for (i = myIndex +1 ; i <project.serverList.size(); i++)
  {
    if (sTopologyInfo[i][myIndex] != CONN) continue;

    newSoc = connectToServer((char *)project.serverList[i].serverName.c_str(),project.serverList[i].gatewayPort,
                    project.serverList[myIndex].minRank,project.serverList[myIndex].maxRank, 
                    SERVICE_SERVER);
    if (newSoc == -1) 
    {
      printf("ERROR. connectAllServers. Can't connect to %s:%d\n",
             project.serverList[i].serverName.c_str(),project.serverList[i].gatewayPort);
      return 0;
    }

    lnk->soc = newSoc;
    lnk->minRank = project.serverList[i].minRank;
    lnk->maxRank = project.serverList[i].maxRank;
    lnk->state = STATE_OPENED;
    lnk->service = SERVICE_SERVER;
    printf("Thread gets all info(link to server). Min %d Max %d\n",
            lnk->minRank,lnk->maxRank);

    links.push_back(*lnk);
  };

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// Create new link and thread that will serve connected socket 'soc'
// if successful returns 1. in case of error returns 0
int Server::createNewLink(int soc)
{
  LinkInfo *lnk = new LinkInfo;
 
  lnk->soc = soc;

  if ( !getLinkInfo(lnk))
  {
    printf("ERROR. createNewLink().Can't retreive new link info\n");      
    return 0;
  };

  lnk->state = STATE_OPENED;
  links.push_back(*lnk);

  printf("New link created... socket: %d, supported rank: ", soc);
  if (lnk->service == SERVICE_CLIENT) printf("%d\n",  lnk->maxRank);
  else
    if (lnk->service == SERVICE_SERVER) printf("Min %d Max %d\n",lnk->minRank, lnk->maxRank);
    else printf("ERROR! Uknown service in a message %d\n",lnk->service);

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
Server *curServ;
void sigCatch(int sig)
{
#ifndef _WIN32
  curServ->terminateServer(SIG_TERMINATE);
  signal(sig, sigCatch);
#endif
};

///////////////////////////////////////////////////////////////////////////////
// Main server cycle
int Server::run()
{
  int i;
  int newSoc;
  fd_set socs;
  struct Packet pack;
  timeval selTime;
  time_t startTime;
  bool wasOper = false;
  int packReceived = 0;
  
  time(&startTime);
#ifndef _WIN32
  curServ = this;
  signal(SIGQUIT, sigCatch);
  signal(SIGTERM, sigCatch);  
  signal(SIGINT, sigCatch);  
#endif  
  while (true)
  {
    if (wasOper) // some operations occured, reset waiting time
    {
      wasOper = false; 
      time(&startTime);
    } 

    fflush(stdout);
    if (time(NULL) - startTime > taskCtrl.serverConfig.jmanLife) // no operations occured during jmanLife time. exit
    {
      printf("ERROR. Server was shutdowned due to a timeout (check file cluster.ini, 'JobmanagerLifetime' parameter). \n");
      terminateServer(SIG_TERMINATE);
      return 0;
    };
    
    FD_ZERO(&socs);
    FD_SET(servSoc, &socs);
    for (i=0; i < links.size(); i++) 
    {
      if (links[i].state == STATE_OPENED)
      {
        FD_SET(links[i].soc, &socs);
      }
    }
    
    selTime.tv_sec = 0;
    selTime.tv_usec = 1000;
//    selTime.tv_usec = 0;

    select(65535, &socs, NULL, NULL, &selTime);
   // select(65535, &socs, NULL, NULL, NULL);
    if (FD_ISSET (servSoc, &socs)) // new connection
    {
      wasOper = true;
      newSoc = acceptConnection(servSoc);

      if (newSoc < 0)
      {
        #ifndef _WIN32
          if (errno == EINTR) continue;
        #endif
      
        if (errno == EBADF) // socket was closed.
        {
          printf("Got connection request. Failed. Connection was suddenly closed. (acceptConnection)\n");
        }
        else printf("ERROR in acceptConnection\n"); 
      }
      else
      {
        if (!createNewLink(newSoc)) 
        {
          printf("Got connection request. Rejected. Unknown service. (acceptConnection)\n");
          closeSocket(newSoc);
        };
      };
    };

    // read and analyze/transfer data from all links
    for (i= 0; i < links.size(); i++) 
    {
      if (FD_ISSET (links[i].soc, &socs)) 
      {
        wasOper = true;
        pack.data = NULL; 

        // receive a new packet
        if (readPacketAlloc(links[i].soc,&pack) == -1)  
        { 
          links[i].state = STATE_NEEDCLOSE;
          printf("ERROR. run(). Can't read packet\n");
          terminateServer(SIG_TERMINATE);
          closeLink(i); 
          break;
        };
//        printf("packet received from rank %d , receiver %d\n", pack.header.sendAddr, pack.header.recvAddr);
//        printf("packet received:");
//        printf("recv: soc=%d,",links[i].soc); showPacketInfo(pack);

        // check for signals from clients
        processSignal(pack.header.signal, i);

        if (pack.header.signal == SIG_NONE)
        {
//          printf("Received packet from soc = %d, address %d, transfering to the client now...\n",
//                    links[i].soc, pack.header.recvAddr);
          packReceived++;
	  if (packReceived % 100 == 0) printf("Received %d packets. ", packReceived);
          links[i].packets.push_back(pack);
        };
	
      }
    } // for. read packets
    writePackets();
  }  // while (true)
}

///////////////////////////////////////////////////////////////////////////////
// Send a signal to all Opened links
int Server::broadcastSignal(int sig)
{
  int i;
  Packet pack;

  for (i =0; i< links.size(); i++)
  {
    if (links[i].state == STATE_OPENED)
    {
      pack.header.signal = sig;
      pack.header.size =0;
      
      if (writePacket(links[i].soc, genPacket(0, 0, 0, MES_P2P, sig, 
                      MES_NONE,0, 0, 0, NULL)) == -1)
      {
        printf("broadcast failed to soc = %d. broadcastSignal().\n", links[i].soc);
        return 0;
      }
    }
  }

  return 1;
}
  
///////////////////////////////////////////////////////////////////////////////
// Find link index for the adress 'addr'
int Server::findSendLink(int addr)
{
  int i;

  for (i =0; i< links.size(); i++)
  {
    if ((addr >= links[i].minRank) &&  
        (addr <= links[i].maxRank)) 
    {
      if (links[i].state == STATE_OPENED) return i;
    }
  } 

  return -1;
}

///////////////////////////////////////////////////////////////////////////////
// check if there are still any messages in the mes. queue
bool Server::isMesQueueEmpty()
{
  int i;

  for (i =0; i < links.size(); i++)
  {
    if (links[i].packets.size() !=0) return false;
  }
  
  return true;
}

///////////////////////////////////////////////////////////////////////////////
void Server::showPacketInfo(Packet &pack)
{
  std::string sig,mes; 
  
  switch (pack.header.mesType)
  {
    case MES_BROADCAST: mes = "MES_BROADCAST"; break;
    case MES_P2P  :mes = "MES_P2P"; break;
  }
  
  switch (pack.header.signal)
  {
    case SIG_NONE           :sig = "SIG_NONE";break;
    case SIG_CLIENT_FINISHED:sig = "SIG_CLIENT_FINISHED"; break;
    case SIG_SERVER_FINISHED:sig = "SIG_SERVER_FINISHED"; break;
    case SIG_TERMINATE      :sig = "SIG_TERMINATE"; break;
  }
  
  printf("PInf: recv=%d, send=%d, Type=%s, Signal=%s,sFunct=%d\n",
             pack.header.recvAddr, pack.header.sendAddr, mes.c_str(), sig.c_str(),
	     pack.header.mesSender);
};
