
#ifndef _SERVER_H
#define _SERVER_H

#pragma warning(disable:4786)

#include <string>
#include <vector>
#include "../Shared/multi_os.h"

#ifdef _WIN32
  #include <windows.h>
#else
  #include <unistd.h>
  #include <sys/socket.h>
  #include <pthread.h>  
#endif

#include "../Shared/projectinfo.h"
#include "../Shared/net.h"
#include "../Shared/common.h"
#include "../Shared/taskctrl.h"

#define CONN        1     // TAG link connected
#define NOT_CONN    0     // TAG link not connected

typedef vector <Packet> PacketList;
///////////////////////////////////////////////////////////////////////////////
enum LinkState { STATE_OPENED, STATE_CLOSED, STATE_NEEDCLOSE};
// Info about a link
struct LinkInfo
{
  int  soc;                          // connected socket
  int  minRank, maxRank;             // min and max rank that are supported by this link
  int  pid;                          // thread PID for this link
  int  service;                      // service type (server, client)

  LinkState state;
  PacketList packets;

  LinkInfo() { state = STATE_CLOSED; };
};

typedef vector <LinkInfo> LinkList;

///////////////////////////////////////////////////////////////////////////////
class Server
{
public:
  Server();
  int findSendLink(int addr);
  int broadcastSignal(int sig);
  int run();
  int init();           // Initialize server.
  void  establishNumeration(); // Establish new numeration in server's list
  void  showJobParams();       // Print job parameters
  int   findThis();       // Locate this server in server's list
  int   connectAllServers();        // Connect all servers at start-up
  void  terminateServer(int sig);   // Terminates server

  int myIndex;           // position of This server in the server's list serverInfo[] 
  ProjectInfo project;   // project description
  int servSoc;           // main server socket.

  string ngPath;
  
  TaskCtrl taskCtrl;

private:
  char *getWorkDir(); // Get current directory
  void  notifyAllServers(int sig);  // Send message to all connected servers with pack.header.mesType = sig
  int   getLinkInfo(LinkInfo *lnk); // Get information about new remote link (server)
  void  closeLink(int index); // close link
  void  processSignal(int sig, int index);
  int   deliverPacket(Packet &pack,  int sendService = SERVICE_CLIENT); // Deliver packet to receiver 
  int   deliverP2Ppacket(Packet &pack);
  int   createNewLink(int soc);     // Create new link and thread that will serve connected socket 'soc'

  void writePackets();
  
  void showPacketInfo(Packet &pack); 

  int sTopologyInfo[MAX_SERVERS][MAX_SERVERS]; // info about server links . topology
  LinkList links; // info about all links of This server
  int serverPort;        // server port

protected:
	bool isMesQueueEmpty();
};

#endif
