
// All common functions 

#include "multi_os.h"

#ifdef _WIN32
  #include <windows.h>
  #include <string>
#else
  #include <unistd.h>
  #include <stdio.h>
  #include <stdarg.h>
  #include <strings.h>
  #include <fcntl.h>
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <sys/wait.h>
  #include <stdlib.h>
#endif

#include "common.h"

// Common functions

extern int DEBUG_LEVEL;

///////////////////////////////////////////////////////////////////////////////
int spawn(char *filename, char *args[], char *outFilename, int &handle)
{
#ifdef _WIN32
#else
  int pid, ret;
  char buf[256];
  char *output = NULL;
  FILE *fp;
  char tmpName[10];
  
  int stdoutSave, stderrSave;
 
  stdoutSave = dup(fileno(stdout));
  stderrSave = dup(fileno(stderr));
  
  getcwd(buf, 256);
  
  strcpy(tmpName, "outXXXXXX");
  if (mktemp(tmpName) == NULL)
  {
    printf("ERROR. Can't create temporary filename for output. Current directory is %s\n", buf);
    return 0;
  };
  
  fp = fopen(tmpName, "w");
  if (fp == NULL)
  {
    printf("ERROR. Can't open temporary file for output. Current directory is %s\n", buf);
    return 0;
  };

  dup2(fileno(fp), fileno(stdout));
  dup2(fileno(fp), fileno(stderr));
  
  pid = fork();
  switch (pid)
  {
    case  0: ret = execv(filename,  args); // child
             if (ret == -1)
             {
               ret = execvp(filename,  args);
               if (ret == -1)
               {
                 dup2(stdoutSave, fileno(stdout));
                 dup2(stderrSave, fileno(stderr));
                 printf("ERROR. TaskCtrl. exec(). Can't run executable '%s'.\nCurrent directory is %s\n", filename, buf);
                 return 0;
               }
             };
    case -1: printf("ERROR. TaskCtrl. exec(). Can't create child process. \n"); // error
             return 0;
    default: break; //parent
  }
  dup2(stdoutSave, fileno(stdout));
  dup2(stderrSave, fileno(stderr));
  usleep(10);
  
  strcpy(outFilename, tmpName);
  handle = pid;
  return 1;
#endif 

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// Perform a lowercase comparison of strings
bool isStrEqual(std::string str1, std::string str2)
{
  if (strcmpi(str1.c_str(), str2.c_str()) == 0) return true;
  else return false;
};

///////////////////////////////////////////////////////////////////////////////
// Suspends the execution of the current thread for a specified interval
// in milliseconds
void delay(int ms)
{
#ifdef _WIN32
  Sleep(ms);
#else
  usleep(ms *1000);
#endif
};

//////////////////////////////////////////////////////////////////////////////
// Writes debug info
void debug(int lev, char *mes,...)
{
//#ifdef DEBUG_LEVEL
  static char str[255];
  int i;
  
  for(i =0; i < lev-1; i++) 
    if (i<10) str[i]=' ';
    else break;
  
  str[i] = 0;
  strcat(str,mes);
  if (lev >= DEBUG_LEVEL) return;

  va_list args;
  va_start(args, mes);
  
  vfprintf(stdout, str, args);
//  fflush(stdout);

//#endif
};

//////////////////////////////////////////////////////////////////////
std::string mergePath(std::string &path, std::string dir, std::string fname)
{
  path = "";
  
  if (dir != "")
  {
    path = dir;
    if ((dir[dir.size()-1] != '/') && (dir[dir.size()-1] != '\\'))
      path+='/';
  }

  if (fname != "") path += fname;
  return path;
};

///////////////////////////////////////////////////////////////////////////////
#ifdef _WIN32
int getRegistryKeys(char **path)
{
  HKEY hKey, keyResult;
  DWORD dwType;
  DWORD dwSize;
  static unsigned char *pBuffer = NULL;

  hKey = HKEY_CURRENT_USER;

  if (RegOpenKeyEx(hKey, "Software", NULL, KEY_READ,&keyResult) == ERROR_SUCCESS)
  {
    RegCloseKey(hKey);
    hKey = keyResult;
    if (RegOpenKeyEx(hKey, "NumGRID", NULL, KEY_READ,&keyResult) == ERROR_SUCCESS)
    {
      RegCloseKey(hKey);
      hKey = keyResult;
      if (RegQueryValueEx(hKey, "Path", NULL, &dwType, NULL, &dwSize) == ERROR_SUCCESS)
      {
        pBuffer = (unsigned char *)malloc(dwSize);
        if (pBuffer == NULL)
        {
          printf("ERROR. getRegistryKeys(). Insufficient memory.\n");
          return 0;
        }
        if (RegQueryValueEx(hKey, "Path", NULL, &dwType, pBuffer, &dwSize) == ERROR_SUCCESS)
        {
			    switch (dwType)
          {  
			      case REG_SZ: *path = (char *)pBuffer; break;
            default: printf("ERROR. Bad registry key type. The key 'HKEY_CURRENT_USER/Software/NumGRID/Path' must be a string. \n"); break;
          }
        }
        else
          printf("Error reading key.\n");

        RegCloseKey(hKey);
//        free(pBuffer);
        return 1;
      }
    }
  }

  path = NULL;
  RegCloseKey(hKey);
  return 0;
};
#endif

//////////////////////////////////////////////////////////////////////
std::string getNgPath()
{
  char *path = NULL;
  static std::string ngPath;

#ifdef _WIN32
  if (getRegistryKeys(&path))
  {
    ngPath = path;
  }
  else
#endif
  {
    ngPath = "";
    if (getenv(NGPATH_ENV) == NULL)
    {
      printf("ERROR. Set environment variable %s to the directory where NumGRID package installed\n", NGPATH_ENV);
      return "";
    }
    else 
      ngPath =  getenv(NGPATH_ENV);
  }
  
  if ((ngPath[ngPath.size() - 1] != '/') &&
      (ngPath[ngPath.size() - 1] != '\\'))
    ngPath+='/';
//printf("----- NGPATH = %s\n", ngPath.c_str());
  return ngPath;
}
