
// This file contains all important constants for all modules

#ifndef _COMMON_H
#define _COMMON_H

#include "multi_os.h"           // Multiple OS (Windows, UNIX) definitions

#include <string>
#include <vector>

#define NG_VERSION 3  // NumGRID version

//////////////////////////////////////////////////////////////////////////////
//        Default file names
#define DEFAULT_JOB_FILENAME     "task.ng"

#define SERVER_CFG_FILE          "cluster.ini"

#define NGPATH_ENV               "NUMGRID" // Environment variable that points to NumGRID directory

//////////////////////////////////////////////////////////////////////////////
#define MAX_DIRECTORY_LENGTH     255 
#define MAX_HOSTNAME_LENGTH      100
#define MAX_STRING_LENGTH        255

#define MAX_SERVERS              10      // Maximum servers(gateways) that are supported
#define THREAD_STACK_SIZE        10000   // STACK size for the server's threads

#define DEFAULT_SERVER_PORT      9000    // Default server port
#define DEFAULT_GATEWAY_PORT     9001    // Default gateway port
#define MIN_PORT_RANGE           1000    // Minimum port value that servers support
#define MAX_PORT_RANGE           30000   // Maximum port value that servers support

#define TIME_WAIT_CLIENT_CON     5       // (millisec.) Server waits this amount of time beetween tries
                                         // to find a new connection (in con list) to
                                         // write message to client

#define TIME_BETWEEN_CONNECT_TRY 1000     // Time to wait between attempts to connect the
                                          // remote host&port (in milliseconds)
#define CONNECT_TRY_TIMES        10       // How many attempts to do (number) 
                                          // Total time to try is CONNECT_TRY_TIMES * TIME_BETWEEN_CONNECT_TRY

#define PROGRAM_ID               519     // program ID

// message ids (signals)
const char SIG_NONE              =  0;   // no signal
const char SIG_CLIENT_FINISHED   =  1;   // client shutdown code (normal)
const char SIG_SERVER_FINISHED   =  2;   // server shutdown code (normal)
const char SIG_TERMINATE         =  3;   // error code (abortive termination)
const char SIG_SYNHRONIZE        =  4;

// signals for User interface
const char SIG_OK              =  0;   // no error
const char SIG_ERROR           =  1;   // error
const char SIG_CONT            =  2;   // error
const char SIG_JOB_RUN_MPI     =  107;
const char SIG_JOB_ALLKILL     =  108;
const char SIG_JOB_RUN         =  100;
const char SIG_JOB_KILL        =  101;
const char SIG_JOB_PUTDATA     =  102;
const char SIG_JOB_GETDATA     =  103;
const char SIG_GET_OUTPUT      =  104;
const char SIG_GET_STATE       =  105;
const char SIG_JOB_BUILD       =  106;


// message types
const char MES_P2P               =  1; // point-to-point message
const char MES_BROADCAST         =  2; // Broadcast mess


// who send this message 
const char MES_NONE             =  0; 
const char MES_MPI_SEND         =  1; 
const char MES_MPI_SENDRECV     =  2; 
const char MES_MPI_BCAST        =  3; 
const char MES_MPI_BARRIER      =  4; 
const char MES_MPI_REDUCE       =  5; 
const char MES_MPI_ALLREDUCE    =  6; 
const char MES_MPI_ALLGATHER    =  7; 

//////////////////////////////////////////////////////////////////////////////
typedef std::vector<std::string> StringList;
typedef std::vector<int> IntList;

//////////////////////////////////////////////////////////////////////////////
//Commonly used functions

int spawn(char *filename, char *args[], char *outFilename, int &handle);
bool isStrEqual(std::string str1, std::string str2);

// Suspends the execution of the current thread for a specified interval
// in milliseconds
void delay(int ms);
void debug(int lev, char *str,...);
std::string mergePath(std::string &path, std::string dir, std::string fname);
std::string getNgPath();


#endif
