
#pragma warning(disable:4786)

#ifdef _WIN32
  #include <io.h>
  #include <direct.h>
#else
  #include <sys/types.h>
  #include <sys/stat.h>
#endif  

#include <openssl/des.h>
#include <openssl/rsa.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/opensslv.h>

# include "esock.h"
# include "net.h"

//////////////////////////////////////////////////////////////////////
ENCSOCK esockConnect(int s)
{
  unsigned char tmp[24];
  Packet pack;

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DES_cblock k1,k2;
#else
  des_cblock k1,k2;
#endif

  RSA *rsa;
  int len;

  _ENCSOCK *es = (_ENCSOCK*)malloc(sizeof(_ENCSOCK));

  if (readPacketAlloc(s, &pack) == -1)
  {
    printf("ERROR. esockConnect. Can't establish security context\n ");
    return 0;
  }

  rsa=0;
  unsigned char *ptr = (unsigned char *)pack.data;

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  rsa=d2i_RSAPublicKey(&rsa,(const unsigned char**)&ptr,pack.header.size);
#else
  rsa=d2i_RSAPublicKey(&rsa,(unsigned char**)&ptr, pack.header.size);
#endif

  if(!rsa)
  {
    printf("ERROR. esockConnect. Can't establish security context\n");
    return 0;
  }

  len=RSA_size(rsa);
  RAND_bytes(tmp,24);
  memcpy(es->iv,tmp,8);
  memcpy(k1,&tmp[8],8);
  memcpy(k2,&tmp[16],8);
  
  if(RSA_public_encrypt(24, tmp, (unsigned char *)pack.data, rsa, RSA_PKCS1_PADDING) <= 0)
  {
    printf("ERROR. Can't encrypt data with public key\n");
    return 0;
  }

  if(writePacket(s,&pack) == -1)
  {
    printf("ERROR. Can't establish security context\n");
    return 0;
  }

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DES_set_key(&k1,&es->ks1);
  DES_set_key(&k2,&es->ks2);
#else
  des_set_key(&k1,es->ks1);
  des_set_key(&k2,es->ks2);
#endif
  es->soc=s;

  return (ENCSOCK)es;
}

//////////////////////////////////////////////////////////////////////
ENCSOCK esockAccept(int s, void *_rsa)
{
  unsigned char buf[1024], *ptr=buf, tmp[24];
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DES_cblock k1,k2;
#else
  des_cblock k1,k2;
#endif
  int len,rd;
  Packet pack;

  RSA *rsa=(RSA*)_rsa;
  _ENCSOCK*es=(_ENCSOCK*)malloc(sizeof(_ENCSOCK));

  if(!rsa)
  {
    printf("ERROR. Can't generate temprorary RSA key\n");
    return 0;
  }

  rd=i2d_RSAPublicKey(rsa,&ptr);
  if(rd <= 0)
  {
   printf("ERROR. Can't convert PEM to DER\n");
   return 0;
  }

  pack.data = (char *)buf;
  pack.header.size = 1024;
  if(!writePacket(s, &pack))
  {
    printf("ERROR. esockAccept. Can't establish security context\n");
    return 0;
  }

  len=RSA_size(rsa);

  if(!readPacket(s, &pack))
  {
    printf("ERROR. esockAccept. Can't establish security context\n");
    return 0;
  }

  rd=RSA_private_decrypt(len, (unsigned char *)pack.data, tmp, rsa, RSA_PKCS1_PADDING);
  if(rd <= 0)
  {
    printf("ERROR. Can't decrypt data with a private key\n");
    return 0;
  }
  if(rd != 24)
  {
    printf("ERROR. Can't decrypt data with private key\n");
    return 0;
  }

  memcpy(es->iv,tmp,8);
  memcpy(k1,&tmp[8],8);
  memcpy(k2,&tmp[16],8);

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DES_set_key(&k1,&es->ks1);
  DES_set_key(&k2,&es->ks2);
#else
  des_set_key(&k1,es->ks1);
  des_set_key(&k2,es->ks2);
#endif

  es->soc=s;
  return (ENCSOCK)es;
}

//////////////////////////////////////////////////////////////////////
//int esockWritePacket(ENCSOCK es, const void *buf, int len)
int esockWritePacket(ENCSOCK es, Packet *pack)
{
  int len2;
  unsigned char *buf2;

  _ENCSOCK *_es=(_ENCSOCK*)es;

  pack->header.dataTypeSize = pack->header.size;

  len2 = pack->header.size;
  if(len2 % 8) len2 += 8 - len2 % 8;
  buf2=(unsigned char*)malloc(len2+8);
  memset(buf2, 0, 8);

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DES_ede2_cbc_encrypt((unsigned char*)pack->data, buf2, pack->header.size, &_es->ks1, &_es->ks2, &_es->iv, 1);
#else
  des_ede2_cbc_encrypt((unsigned char*)pack->data, buf2, pack->header.size, _es->ks1, _es->ks2, &_es->iv, 1);
#endif

  pack->data = (char *)buf2;
  pack->header.size = len2;

  if(writePacket(_es->soc,pack) == -1)
  {
    printf("ERROR. Can't write packet\n");
    return 0;
  }
  free(buf2);
  return 1;
}

//////////////////////////////////////////////////////////////////////
int esockReadPacket(ENCSOCK es, Packet *pack)
{
  unsigned char *buf2;

  _ENCSOCK*_es=(_ENCSOCK*)es;

  if(readPacket(_es->soc,pack) == -1)
  {
    printf("ERROR. Can't read packet\n");
    return 0;
  }

  buf2 = (unsigned char*)malloc(pack->header.size);

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DES_ede2_cbc_encrypt((unsigned char *)pack->data, buf2, pack->header.size, &_es->ks1, &_es->ks2, &_es->iv, 0);
#else
  des_ede2_cbc_encrypt((unsigned char *)pack->data, buf2, pack->header.size, _es->ks1, _es->ks2, &_es->iv, 0);
#endif

  pack->header.size = pack->header.dataTypeSize;
  memcpy(pack->data, buf2, pack->header.size);
  free(buf2);
  return 1;
}

//////////////////////////////////////////////////////////////////////
int esockReadPacketAlloc(ENCSOCK es, Packet *pack)
{
  unsigned char *buf2;

  _ENCSOCK*_es=(_ENCSOCK*)es;

  if(readPacketAlloc(_es->soc,pack) == -1)
  {
    printf("ERROR. Can't read packet\n");
    return 0;
  }

  buf2 = (unsigned char*)malloc(pack->header.size + 8);
  memset(buf2, 0, 8);

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DES_ede2_cbc_encrypt((unsigned char *)pack->data, buf2, pack->header.size, &_es->ks1, &_es->ks2, &_es->iv, 0);
#else
  des_ede2_cbc_encrypt((unsigned char *)pack->data, buf2, pack->header.size, _es->ks1, _es->ks2, &_es->iv, 0);
#endif

  pack->header.size = pack->header.dataTypeSize;
  memcpy(pack->data, buf2, pack->header.size);
  free(buf2);
  return 1;
}

//////////////////////////////////////////////////////////////////////
int writeMessage(ENCSOCK es, char *mes, int sig)
{
  Packet pack;
  pack.header.signal       = sig;
  pack.header.size         = strlen(mes);
  pack.data                = mes;

  return esockWritePacket(es, &pack);
};

//////////////////////////////////////////////////////////////////////
int readMessage(ENCSOCK es, char **mes, int &len, int &sig)
{
  Packet pack;

  if (!esockReadPacketAlloc(es, &pack)) return 0;
  *mes = pack.data;
  len = pack.header.size;
  sig = pack.header.signal;

  return 1;
};

//////////////////////////////////////////////////////////////////////
int writeFile(ENCSOCK es, std::string filename, std::string dir)
{
  Packet pack;
  FILE *fp;
  int len;
  int i;
  std::string path="";

  if (filename == "")
  {
    printf ("ERROR. writeFile. Empty filename passed = ''\n");
    return 0;
  }

  mergePath(path, dir, filename);

  fp = fopen(path.c_str(), "r");
  if (fp == NULL) 
  {
    printf("ERROR. Can't open file %s for transfer\n", filename.c_str());
    return 0;
  };

#ifdef _WIN32
  len = _filelength(_fileno(fp));
#else
  struct stat st;
  if (stat(path.c_str(), &st) ==-1)
  {
    printf ("Error in system call -'stat()'. Code  %d\n", errno);
    return 0;
  }
  len = st.st_size;
#endif

  pack.header.size = len + filename.size() + 2 ;
  char *buf = (char *)malloc(len + filename.size() + 2);
  i =0;

  strcpy(buf, filename.c_str());
  i+= filename.size();
  buf[i++] = 0;
  while (!feof(fp))
  {
    buf[i++] = (char)fgetc(fp);
  }

  pack.data = buf;
  pack.header.size = i - 1 ;
  
  if (!esockWritePacket(es, &pack)) return 0;

  fclose(fp);
  return 1;
};

//////////////////////////////////////////////////////////////////////
// if filename = NULL (default), then save file with the name received,
// else save it as 'filename'
int readFile(ENCSOCK es, std::string filename, std::string dir)
{
  Packet pack;
  FILE *fp;
  std::string path="";

  if (!esockReadPacketAlloc(es, &pack)) return 0;

  char *pos = strchr(pack.data, 0);
  if (pos == NULL)
  {
    printf("ERROR. Can't get 'filename' from the received packet\n");
    return 0;
  };

  if (filename != "") mergePath(path, dir, filename);
  else mergePath(path, dir, pack.data);

  fp = fopen(path.c_str(), "w");

  if (fp == NULL) 
  {
    if (dir!= "")
    {
     #ifdef _WIN32
       if (mkdir(dir.c_str()) != -1)
     #else
       if (mkdir(dir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH) != -1)
     #endif  
         fp = fopen(path.c_str(), "w");
    }

    if (fp == NULL) 
    {
      printf("ERROR. Can't open file %s for writing\n", filename.c_str());
      return 0;
    }
  };

  for (int i=pos - pack.data + 1; i < pack.header.size; i++)
    fputc(pack.data[i], fp);

  fclose(fp);
  return 1;
};

//////////////////////////////////////////////////////////////////////
int writeFileList(ENCSOCK es, StringList lst, std::string dir)
{
  if (!writeMessage(es, "", lst.size())) return 0;

  for (int i=0; i <lst.size(); i++)
    if (!writeFile(es, (char *)lst[i].c_str(), dir)) return 0;

  return 1;
};

//////////////////////////////////////////////////////////////////////
int readFileList(ENCSOCK es, std::string dir)
{
  int len, sig;
  char *mes;
  if (!readMessage(es, &mes, len, sig)) return 0;

  for (int i=0; i <sig; i++)
    if (!readFile(es, "", dir)) return 0;

  return 1;
};

//////////////////////////////////////////////////////////////////////
