
// Low level network functions

#ifndef _NET_H
#define _NET_H

#include "common.h"

#define SERVICE_CLIENT   0
#define SERVICE_SERVER   1

// data structures 
///////////////////////////////////////////////////////////////////////////////

// packet header
struct PacketHeader
{
  int recvAddr;    // address of receiver 
  int sendAddr;    // address of sender 
  int tag;         // packet tag (contains MPI tag)
  char mesType;    // type of a message (system value)
  char signal;     // system signal
  char mesSender;  // who sent (function), for error checking (system value)

  int size;        // size of packet body (*dataTypeSize)
  int dataType;    //
  int dataTypeSize;
};

// main net packet 
struct Packet
{
  struct PacketHeader header;  // header
  char *data ;                 // packet body

  Packet() {data = NULL; };
};

// connection info for servers. send at start-up
struct ConnectInfo
{
  int id;                       // program ID
  int ngVersion;                // NumGRID version id
  int service;                  // client or server
  int minRank,maxRank;          // min and max ranks that server supports
};

// functions
///////////////////////////////////////////////////////////////////////////////

// Call to initialize network library. Must be called before
// using all other network functions
int initNetwork();

// Call to finalize network library. Must be called at the end
void closeNetwork();


// client functions ----------------------------------


// Establish connection to numGRID server
// min and max ranks must be passed as arguments
// service can be  SERVICE_CLIENT or SERVICE_SERVER  
// returns connected socket or -1 in case of error
int connectToServer(char *server, int port, int min, int max, int service);

// server functions ----------------------------------

// Wait for new connection. Returns new socket or -1 in case of error
int acceptConnection(int s);

// Make and bind server socket on port. Returns connected socket
// or -1 in case of error
int makeServerSocket(unsigned short portnum);

// io functions --------------------------------------

// Reads net packet into struct p from the socket . memory MUST be allocated
// before. Returns 1 on success or -1 in case of error
int readPacket(int soc, struct Packet *p);

// Reads packet into struct p from the socket. memory is allocated dynamically.
// later you should free this memory manually using free.
// returns 1 on success or -1 in case of error
int readPacketAlloc(int soc, struct Packet *p);

// Writes net packet from struct p into the socket.
// returns 1 on success or -1 in case of error
int writePacket(int soc, struct Packet *p);

// low level functions (never called directly) --------

// Connect to host and port. Returns connected socket
// or -1 in case of error
int callSocket(char *hostname, unsigned short portnum);

// Reads size bytes from connected socket into buf (must be allocated before!)
// returns number bytes read or -1 in case of error
int readData(int soc, char *buf, int size);
// Writes size bytes to connected socket from buf (must be allocated before!)
// returns number bytes written or -1 in case of error
int writeData(int soc, char *buf, int size);

// Carefully closes socket soc
void closeSocket(int soc);
// common functions -----------------------------------

// Shows last socket error. mes can be used to show function name where
// socket info occured
void showSocketError(char *mes = "");

// fills the packet fields and return a pointer to static packet (for GMPI)
Packet *genPacket(int recv, int send, int tag, char mesType, char signal, char mesSender, 
                  int size, int dataType, int dataTypeSize, char *data);

#endif
