// ProjectInfo.h: interface for the ProjectInfo class.
//
//////////////////////////////////////////////////////////////////////
// Read, parse and keep all information about the project
//

#ifndef _PROJECTINFO_H
#define _PROJECTINFO_H

#include "scripts.h"
#include "multi_os.h"

// Error codes
#define ERR_WRONG_VALUE      200
#define ERR_WRONG_PARAM      201
#define ERR_WRONG_RUNTYPE    202
#define ERR_UNKNOWN_IDENT    203
#define ERR_UNKNOWN_VAR_NAME 204
#define ERR_BAD_PORT         205
#define ERR_INVALID_NP       206

//////////////////////////////////////////////////////////////////////
// Keeps all info about single server
class ServerInfo
{
public:
  void clear();

  string blockName;
  string serverName;
  int serverPort;
  string internalHostName;
  int gatewayPort;

  string username;
  string password;

  int nProcesses;

  string sourceDir;
  string inputDir;
  string outputDir;

  string workDir;

  bool needRunTask;
  string executable;
  StringList execArgs;
  StringList execEnv;

  string makefile;
  StringList sourceFiles;
  StringList inputFiles;
  StringList outputFiles;

  int  minRank,maxRank;   // min and max rank that are supported by this link
};

//////////////////////////////////////////////////////////////////////
// Keeps info about all servers. Allows to load info from the text file
class ProjectInfo  
{
public:

  ProjectInfo();
  virtual ~ProjectInfo();
  
  int LoadInfo(char *filename);        // load info from the file
  char *getErrorText();                // Get last error description 

  vector <ServerInfo> serverList;      // main list of all servers

  int errCode;                         // error code is kept here
  std::string projectFile;  

private:
  int rebuildBlocks();
  int getVarCode(string varName);                  // Find id for the keyword
  int parseVar(NGVarInfo &var,  ServerInfo &sInfo);  // Insert new variable's data into ServerInfo
  int checkParams(void);                           // Check info in the project file
};

#endif 
