
// Scripts functions

#pragma warning(disable:4786)

#include <string.h>

#include "scripts.h"
#include "multi_os.h"
//#include "../Shared/common.h"

///////////////////////////////////////////////////////////////////////////////
// NGVarInfo class methods

NGVarInfo::NGVarInfo()
{
  name ="";
  blockName ="";
}

///////////////////////////////////////////////////////////////////////////////
// NGScript class methods
///////////////////////////////////////////////////////////////////////////////
NGScript:: NGScript()
{
  fp_ = NULL;
  curLine_ = 1;

  commentStr_ = COMMENT_DEFAULT;
  brackSym_ = BRACKET_DEFAULT;
  separSym_ = SEPARATOR_DEFAULT;
  assignSym_ = ASSIGN_DEFAULT;
  blockStartStr_ = BLOCK_START_DEFAULT;
  blockEndStr_ = BLOCK_END_DEFAULT;

  separatorFound = false;
};

///////////////////////////////////////////////////////////////////////////////
NGScript::~NGScript()
{
  if (fp_ != NULL) fclose(fp_);
};

///////////////////////////////////////////////////////////////////////////////
// Open script file
int NGScript::open(char *filename)
{
  fp_ = fopen(filename,"rt");
  if (fp_ == NULL) return 0;
  else return 1;
};

///////////////////////////////////////////////////////////////////////////////
// Read next lexem from text file
// returns pointer to lexem(temporary) or NULL in case of error
// lexems are: strings  & end of string(10)
string NGScript::getLexem()
{
  string str;
  int ch;
  bool brackStart= false;

  if (fp_ == NULL) return "";
  separatorFound = false;
  bracketFound = false;

  do   // skip empty chars
  {
    ch = fgetc(fp_); 
    if (ch == END_STRING)
    {
      curLine_++;
      break;
    }
  }
  while ((ch == SPACE_SYM)  || (ch == RETURN_SYM) || (ch==END_STRING) || (ch==TAB_SYM));

  if (ch == EOF)
  {
    errCode = ERR_END_OF_FILE;
    return "";
  }

  if (ch == brackSym_) // check for bracket 
  {
    brackStart = true; 
    bracketFound = true;
    ch = fgetc(fp_);
    if (ch == EOF) 
    {
      errCode = ERR_SCRIPT_CLOSEBRACK_NOTFOUND;
      return "";
    }
  }

  while (true) // main read cycle
  {
    if (ch == brackSym_) break;

    str+= ch;

    ch = fgetc(fp_);

    if ((str[0] == separSym_) || (str[0] == assignSym_) || (str[0] == END_STRING))
      if (!brackStart) 
      {
        separatorFound = true;
        break;
      }

    if ((ch == RETURN_SYM) || (ch==END_STRING) || (ch==EOF)) // if in "..." show error
    {
      if (brackStart)
      {
        errCode = ERR_SCRIPT_CLOSEBRACK_NOTFOUND;
        return "";
      }

      break;
    }

    if ((ch == TAB_SYM) || (ch==SPACE_SYM) || (ch==separSym_) || (ch==assignSym_)) // if not in "..." break
      if (!brackStart) break;

  }  // while

  if (!brackStart) ungetc(ch, fp_); // keep last symbol in the buffer for next lexem read

  return str;
};

///////////////////////////////////////////////////////////////////////////////
// Read next string from text file
// returns pointer to string(temporary) or NULL in case of error
string NGScript::getString()
{
  static string str;
  char ch;

  if (fp_ == NULL) return "";

  str = "";
  curLine_++;

  ch = fgetc(fp_); 
  while ((ch != END_STRING) && (ch != RETURN_SYM) && (ch!=EOF))
  {
    str+=ch;
    ch = fgetc(fp_); 
  }

  //fgets(buf,255,fp_);

  return str;
};

///////////////////////////////////////////////////////////////////////////////
// Check if str is a comment
bool NGScript::isComment(string str)
{
  if (strstr(str.c_str(),commentStr_.c_str()) == NULL) return false;
  else return true;
};

///////////////////////////////////////////////////////////////////////////////
// Check if str is a block start
bool NGScript::isBlockStart(string str)
{
  if (strcmpi(str.c_str(),blockStartStr_.c_str()) != 0) return false;
  else return true;
};

///////////////////////////////////////////////////////////////////////////////
// Check if str is a block end
bool NGScript::isBlockEnd(string str)
{
  if (strcmpi(str.c_str(),blockEndStr_.c_str()) != 0) return false;
  else return true;
};

///////////////////////////////////////////////////////////////////////////////
// Check if str is a separator
bool NGScript::isSeparator(string str)
{
  if (str.length() != 1) return false;
  if (strchr(str.c_str(),separSym_) != NULL) return true;
  if (strchr(str.c_str(),10) != NULL) return true;
  return false;
};

///////////////////////////////////////////////////////////////////////////////
// Check if str is a assignment
bool NGScript::isAssign(string str)
{
  if (strchr(str.c_str(),assignSym_) == NULL) return false;
  else return true;
};

///////////////////////////////////////////////////////////////////////////////
// Return current line in a file
int NGScript::getCurLine()
{
  return curLine_;
};

///////////////////////////////////////////////////////////////////////////////
// set new lexem rules
void NGScript::setLexSyntax(char *commentStr,char brackSym, char separSym, char assignSym)
{
  commentStr_ =  commentStr;
  brackSym_ = brackSym;
  separSym_ = separSym;
  assignSym_ = assignSym;
}

///////////////////////////////////////////////////////////////////////////////
// set new syntax
void NGScript::setStructSyntax(char *blockStartStr, char *blockEndStr)
{
  blockStartStr_ = blockStartStr;
  blockEndStr_ = blockEndStr;
}

///////////////////////////////////////////////////////////////////////////////
// Read variable. variable = <VarName> + param1, param2, ....
NGVarInfo *NGScript::readVar()
{
  string str, lex1;

  var.name = "";
  var.params.clear();

  blockChanged = false;
  do
  {
    str = getClearLexem();
    if (errCode == ERR_END_OF_FILE) return NULL;
    if (str.empty()) // read first lexem
    {
      errCode = ERR_SCRIPT_NO_VAR_FOUND;
      return  NULL;
    }
  }
  while (str[0] == END_STRING); // skip empty end of strings

  if (isBlockEnd(str))  // check block end "}" , get vars
  {
    if (str.empty()) 
    {
      errCode = ERR_SCRIPT_ILLEGAL_BLOCKEND;
      return NULL;
    };
    var.blockName =  "";
    return readVar();
  };

  lex1 = str;  // read first param
  do
  {
    str = getClearLexem();

    if (str.empty()) // read first lexem
    {
      errCode = ERR_SCRIPT_NO_PARAM_FOUND;
      return  NULL;
    }
  }
  while (str[0] == END_STRING); // skip empty end of strings

  if (strcmpi(str.c_str(),blockStartStr_.c_str()) == 0) // block beginning found
  {
    if (readVar() == NULL) return NULL;
    blockChanged = true;
    var.blockName = lex1;
    return &var;
  }
  else
  if (strchr(str.c_str(),assignSym_) != NULL)  // "varName = " found
  {
    var.name = lex1;
    if (!readParamList()) return NULL;
  }
  else 
  {
    errCode = ERR_SCRIPT_BL_OR_ASS_EXPECTED;
    return  NULL;
  }

  return &var;
};

///////////////////////////////////////////////////////////////////////////////
// Reads variable parameters from the file
bool NGScript::readParamList()
{
  string str;

  while (true)
  {
    str = getLexem();

    if (!bracketFound)
    {
      if (var.params.empty())
      {
        if ((str.empty()) || isAssign(str) || isBlockStart(str) || 
             isBlockEnd(str) || isSeparator(str)|| isComment(str))
        {
          errCode = ERR_SCRIPT_NO_PARAM_FOUND;
          return false;
        }
      }

      if (str.empty()) return true;
      if (str[0] == END_STRING) return true;
      if (isComment(str)) 
      {
        getString();
        return true;
      }

      if (isSeparator(str)) continue;
    }

    if (str[0] == END_STRING) return true;
    var.params.insert(var.params.end(), str);
  }
}

///////////////////////////////////////////////////////////////////////////////
// Get last error description
char * NGScript::getErrorText()
{
  switch (errCode)
  {
    case ERR_SCRIPT_CLOSEBRACK_NOTFOUND: return "Closing bracket expected\n";
    case ERR_SCRIPT_NO_VAR_FOUND: return "Variable expected\n";
    case ERR_SCRIPT_NO_PARAM_FOUND: return "Parameter expected\n";
    case ERR_SCRIPT_ILLEGAL_BLOCKEND: return "Illegal closing bracket\n";
    case ERR_SCRIPT_BL_OR_ASS_EXPECTED: return "Opening block or assignment expected\n";

    default: return "No error occured\n";
  };
}

///////////////////////////////////////////////////////////////////////////////
// Get next lexem , skipping comments
string NGScript::getClearLexem()
{
  string str;

  while (true)
  {
    str = getLexem();
    if (str.empty()) // read first lexem
      return "";

    if (!isComment(str)) return str;

    getString();
  }
}
