/* Scripts functions

Supported scripts syntax

Common block
...
Block1_Name
{

  var = parameter1, ..., parameterN (or var = parameter1 parameter2 ... parameterN)
   // separators are ' ' or ,
   // Comment line
}
...
BlockN_Name
{
...
}

*/
#ifndef _SCRIPTS_H
#define _SCRIPTS_H

#include "multi_os.h"
#include "common.h"

#ifdef _WIN32
  #pragma warning(disable:4786)
#endif

#include <stdio.h>
#include <vector>
#include <string>

using namespace std;

// default script's syntax
const char COMMENT_DEFAULT[]     = "#";
const char BLOCK_START_DEFAULT[] = "{";
const char BLOCK_END_DEFAULT[]   = "}";
const char BRACKET_DEFAULT       = '"' ;
const char SEPARATOR_DEFAULT     = ',' ;
const char ASSIGN_DEFAULT        = '=' ;

const char END_STRING            = 10; // on UNIX & WINDOWS this symbol marks the end of string
const char RETURN_SYM            = 13; // WINDOWS only. 10 13 marks the end of string
const char SPACE_SYM             = ' ';
const char TAB_SYM               = 9;

// lexem errors
#define ERR_SCRIPT_CLOSEBRACK_NOTFOUND   101
#define ERR_END_OF_FILE                  102
// syntax errors
#define ERR_SCRIPT_NO_VAR_FOUND          110
#define ERR_SCRIPT_NO_PARAM_FOUND        111
#define ERR_SCRIPT_ILLEGAL_BLOCKEND      112
#define ERR_SCRIPT_BL_OR_ASS_EXPECTED    113

//typedef vector<string> StringList;

///////////////////////////////////////////////////////////////////////////////
// Variable representation
class NGVarInfo
{
public:
  NGVarInfo();

  string name;          // var name
  string blockName;     // var's block name
  StringList params;    // param array(list)
};

///////////////////////////////////////////////////////////////////////////////
// Contains all methods for script files reading
class NGScript
{
public:

  NGScript();
  ~NGScript();

  // set diffirent syntax
  void setStructSyntax(char *blockStartStr, char *blockEndStr);
  void setLexSyntax(char *commentStr,char brackSym, char separSym, char assignSym);

  int   open(char *filename);  // Open script file
  string getLexem();           // Read next lexem from text file
  string getClearLexem();      // Get next lexem , skipping comments
  string getString();          // Read next string from text file
  int   getCurLine();          // Return current line in a file
  NGVarInfo *readVar();          // read variable. variable = <VarName> + param1, param2, ....
  char * getErrorText();       // Get last error description

  bool isComment(string str);    // Check if str is a comment
  bool isBlockStart(string str); // Check if str is a block start
  bool isBlockEnd(string str);   // Check if str is a block end
  bool isSeparator(string str);  // Check if str is a separator
  bool isAssign(string str);     // Check if str is a assignment '='

  int  errCode;                 // error code is kept here
  bool separatorFound;          // was separator found during last read operation?
  bool blockChanged;            // block was changes during last operation
  bool bracketFound;            // brackets were found during last read operation

  NGVarInfo var;                   // last read variable is kept here

private:
  bool readParamList();          // reads variable parameters from the file
  FILE *fp_;                     // file descriptor 
  int curLine_;                  // current line
  string commentStr_;            // current syntax for this object
  string blockStartStr_;         // (for default values see above)
  string blockEndStr_;           //

  char brackSym_;                      //
  char separSym_;                      //
  char assignSym_;                     // 
};
///////////////////////////////////////////////////////////////////////////////

#endif
