
#pragma warning(disable:4786)

#include "common.h"
#include "serverconfig.h"
#include "serversyntax.h"

///////////////////////////////////////////////////////////////////////////////
ServerConfig::ServerConfig()
{
  clear();
};

///////////////////////////////////////////////////////////////////////////////
// load info from the file
int ServerConfig::loadInfo(char * filename)
{
  NGScript s;
  string blockName = "";

  if (!s.open(filename)) 
  {
    printf("Can't open file with Nodes configuration (%s)\n", filename);
    return 0;
  }

  while (true)
  {
    if (s.readVar() == NULL) // error
    {
      if (s.errCode == ERR_END_OF_FILE) break;

      printf("ERROR in line %d, %s\n",s.getCurLine(), s.getErrorText());
      return 0;
    };

    if (!parseVar(s.var))
    {
      printf("ERROR in line %d, %s\n",s.getCurLine() -1 , getErrorText());
      return 0;
    };
  }

  return 1;
}

///////////////////////////////////////////////////////////////////////////////
void ServerConfig::clear(void)
{
  queue = QUEUE_MPICH;
  runScript = "run.sh";
  addScript = "node.sh";
  paramsPassMethod = PARAMS_ENV;
  jmanLife = 86400; // a day
}

///////////////////////////////////////////////////////////////////////////////
int ServerConfig::parseVar(NGVarInfo &var)
{
  int code;
  int parCode;

  code = getVarCode(var.name);
  if (code == PAR_ERROR) 
  {
    errCode = ERR_UNKNOWN_IDENT;
    return 0;
  };

  switch(code)
  {
    case PAR_QUEUE          :parCode = getVarCode(var.params[0]);
                             switch (parCode)
                             {
                               case PAR_QUEUE_MPICH : queue = QUEUE_MPICH; break;
                               case PAR_QUEUE_LAMMPI : queue = QUEUE_LAMMPI; break;
                               case PAR_QUEUE_SGE : queue = QUEUE_SGE; break;
                               case PAR_QUEUE_CUSTOM : queue = QUEUE_CUSTOM; break;
                               case PAR_QUEUE_PBS : queue = QUEUE_PBS; break;
                               default            : errCode = ERR_WRONG_QUEUETYPE; return 0;
                             };
                             break;
    case PAR_PASS_TYPE      :parCode = getVarCode(var.params[0]);
                             switch (parCode)
                             {
                               case PAR_PASS_ENV : paramsPassMethod = PARAMS_ENV; break;
                               case PAR_PASS_ARG : paramsPassMethod = PARAMS_ARG; break;
                               default           : errCode = ERR_WRONG_PARAMSTYPE; return 0;
                             };
                             break;
    case PAR_RUN_SCRIPT     :runScript = var.params[0]; break;
    case PAR_ADD_SCRIPT     :addScript = var.params[0]; break;
    case PAR_JMAN_LIFE      :jmanLife = atoi(var.params[0].c_str()); break;

    default: errCode = ERR_UNKNOWN_VAR_NAME; return 0;
  };

  return 1;
}

//////////////////////////////////////////////////////////////////////
int ServerConfig::getVarCode(string varName)
{
  int i;

  for (i =0; i<KEYWORDS_NUM; i++)
    if (isStrEqual(varName, SCRIPT_KEYWORDS[i].name)) return SCRIPT_KEYWORDS[i].id;

  return PAR_ERROR;
}

//////////////////////////////////////////////////////////////////////
char *ServerConfig::getErrorText()
{
  switch (errCode) 
  {
    case ERR_WRONG_VALUE      : return "Incorrect numeric value\n";
    case ERR_WRONG_QUEUETYPE  : return "Incorrect queue type. 'SGE', 'MPICH', 'LAMMPI', 'Custom' expected\n";
    case ERR_UNKNOWN_IDENT    : return "Unknown identifier\n";
    case ERR_UNKNOWN_VAR_NAME : return "Unknown variable\n";
    case ERR_WRONG_PARAMSTYPE : return "Unknown type for ParamsPassMethod variable. 'Env' or 'Arg' expected\n";
    default: return "No error occured\n";
  }
};
