// TaskCtrl.cpp: implementation of the TaskCtrl class.
//
///////////////////////////////////////////////////////////////////////////////

#pragma warning(disable:4786)

#include "taskctrl.h"
#include "common.h"

#ifdef _WIN32
  #include <windows.h>
  #include <time.h>
  #include <direct.h>
  #include <io.h>
  #include <process.h>
#else
  #include <unistd.h>
  #include <errno.h>
  #include <fcntl.h>
  #include <sys/stat.h>
  #include <sys/types.h>
#endif

#include "common.h"

const char *NG_EXEC       = "<NG_Exec>";
const char *NG_EXECARGS   = "<NG_ExecArgs>";
const char *NG_EXECENV    = "<NG_ExecEnv>";
const char *NG_NP         = "<NG_Np>";

const char *NG_NEW_RUN_SH_NAME   = "run_ng_tmp.sh";

///////////////////////////////////////////////////////////////////////////////
// Exec another process
int exec(char *filename, char *args[])
{
#ifdef _WIN32
#else
  int id;
  int ret;
  char buf[255];

  getcwd(buf, 255);
  id = fork();
  
  switch (id)
  {
    case  0: break; break; // child
    case -1: printf("ERROR. TaskCtrl. exec(). Can't create child process. \n"); // error
             return 0;
    default: return 1; //parent
  }

  ret = execv(filename,  args);
  if (ret == -1)
  {
    ret = execvp(filename,  args);
    if (ret == -1)
    {
      printf("ERROR. TaskCtrl. exec(). Can't run executable '%s'.\nCurrent directory is %s\n", filename, buf);
      return 0;
    }
  };
#endif 

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// TaskCtrl class
///////////////////////////////////////////////////////////////////////////////
TaskCtrl::TaskCtrl()
{
}

///////////////////////////////////////////////////////////////////////////////
TaskCtrl::~TaskCtrl()
{
}

///////////////////////////////////////////////////////////////////////////////
void TaskCtrl::init(ProjectInfo info, int myIndex, string ngPath)
{
  myIndex_ = myIndex;
  info_ = info;
  ngPath_= ngPath;
}

///////////////////////////////////////////////////////////////////////////////
int TaskCtrl::runMPI()
{
  static char envParStr1[255];
  char npStr[255], exeStr[255],indStr[255];
  string envParams;
  string envKey;
  char *args[255];
  int i;

  sprintf(npStr,"%i", info_.serverList[myIndex_].nProcesses);
  strncpy(exeStr, info_.serverList[myIndex_].executable.c_str(), 255);
  sprintf(indStr,"%i", myIndex_);

   // set environ vars
  if (serverConfig.queue == QUEUE_MPICH) //MPICH
  {
    envParams = "\"";
    for (i = 0; i < info_.serverList[myIndex_].execEnv.size(); i++)
    {
      envParams+=info_.serverList[myIndex_].execEnv[i];
      envParams+="|";
    }

    envParams += "_GMPI_SPEC_FILE=";
    envParams += info_.projectFile.c_str();
    envParams += "|_GMPI_SPEC_NUM=";
    envParams += indStr;envParams += "\""; 
    envKey = "-env";
  }
  else // LAM-MPI
  {
    envParams = "";
    for (i = 0; i < info_.serverList[myIndex_].execEnv.size(); i++)
    {
      putenv((char *)info_.serverList[myIndex_].execEnv[i].c_str());
      
      int pos1 = info_.serverList[myIndex_].execEnv[i].find_first_of("=");
      int pos2 = info_.serverList[myIndex_].execEnv[i].find_first_of(" ");
      if ((pos2 >0) && (pos2 < pos1)) pos1 = pos2;
      
      strncpy(envParStr1, info_.serverList[myIndex_].execEnv[i].c_str(),pos1);
      printf("%s,%s,\n",envParStr1, info_.serverList[myIndex_].execEnv[i].c_str());
      envParams+=envParStr1;
      envParams+=",";
    }
    
    envParams += "_GMPI_SPEC_FILE,_GMPI_SPEC_NUM";
    envKey = "-x";
  }

  //set args
  if (serverConfig.paramsPassMethod == PARAMS_ENV)
  {
    args[0] = "mpirun";
    args[1] = "-np";
    args[2] = npStr;
    args[3] = (char *)envKey.c_str();
    args[4] = (char *)envParams.c_str();
    args[5] = exeStr;

    if (info_.serverList[myIndex_].execArgs.size() > 200)
    {
      printf("ERROR. TaskCtrl. runMPI(). Too many command line arguments (>200)\n");
      return 0;
    };
    for (i = 0; i < info_.serverList[myIndex_].execArgs.size(); i++)
      args[6+i] = (char *)info_.serverList[myIndex_].execArgs[i].c_str();

    args[6+i] = NULL;
  }
  else
  {
    if (info_.serverList[myIndex_].execArgs.size() >0)
    {
      printf("ERROR. TaskCtrl. runMPI(). You can't combine ParamsPassMethod=PARAMS_ARG and passing arguments to the program through configuration file(execArgs=...). Try to set ParamsPassMethod = PARAMS_ENV in cluster.ini file in NumGRID home directory\n");
      return 0;
    };
    args[0] = "mpirun";
    args[1] = "-np";
    args[2] = npStr;
    args[3] = exeStr;;
    args[4] = (char *)info_.projectFile.c_str();
    args[5] = indStr;
    args[6] = NULL;
  };
  
#ifdef _WIN32
  int err;
  err = _spawnvp(_P_NOWAIT , "mpirun", args);

  if (err == -1)
  {
    switch (errno)
    {
      case ENOENT :printf("ERROR. TaskCtrl. runMPI(). No 'mpirun' module found \n");
                   return 0;
      default     :printf("ERROR. TaskCtrl. runMPI(). Error in spawnlp \n");
		               return 0;
    };
  };

#else // UNIX version
  if (!exec("mpirun", args)) return 0;
#endif

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// support only ARG pass type. no env support now.
int TaskCtrl::runSGE()
{
  char buf[255];
  char *args[6];

  // prepare SGE script
  FILE *fp = fopen(genRunScriptName(),"wt");

  if (fp == NULL)
  {
    printf("ERROR. TaskCtrl().runSGE. Can't create run script for SGE queue\n");
    return 0;
  }

  fprintf(fp, "#!/bin/sh\n");
  fprintf(fp, "#$ -cwd\n");
  fprintf(fp, "mpirun -np $NSLOTS -machinefile $TMPDIR/machines ./%s %s %d\n",
          info_.serverList[myIndex_].executable.c_str(), info_.projectFile.c_str(),
	  myIndex_);
  fclose(fp);

  // run
  args[0] = "qsub";
  args[1] =  "-pe";
  args[2] = "mpich";
  sprintf(buf,"%d", info_.serverList[myIndex_].nProcesses);  
  args[3] = buf;
  args[4] = genRunScriptName();
  args[5] = NULL;
  
  if (!exec("qsub", args)) return 0;

  return 1;
}

///////////////////////////////////////////////////////////////////////////////
int TaskCtrl::runCustom()
{
  char *args[6];

  args[0] = NULL;

  if (!makeAllScripts()) return 0;
  if (!exec((char *)NG_NEW_RUN_SH_NAME, args)) return 0;

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
/*int TaskCtrl::runSimple()
{
  char *args[6];

  args[0] = NULL;

  if (!exec((char *)info_.serverList[myIndex_].executable.c_str(), args)) return 0;

  return 1;
};*/

///////////////////////////////////////////////////////////////////////////////
int TaskCtrl::run()
{
  static char envParStr1[255];
  static char envParStr2[255];  
  string envParStr;
  char tmpBuf[255];
  FILE *fp;

  if (serverConfig.paramsPassMethod == PARAMS_ENV)
  {
      envParStr = "_GMPI_SPEC_FILE="; envParStr+=info_.projectFile.c_str();
      strncpy(envParStr1, envParStr.c_str(), 255);
      putenv(envParStr1);
      
      envParStr = "_GMPI_SPEC_NUM="; 
      sprintf(tmpBuf,"%i", myIndex_);
      envParStr+=tmpBuf;
      strncpy(envParStr2, envParStr.c_str(), 255);
      putenv(envParStr2);
  }
  
  // check exec file existance
  if ((fp = fopen(info_.serverList[myIndex_].executable.c_str(), "r")) == NULL)
  {
    printf("ERROR. No executable of application program was found (%s). Check that you built or copied it to the cluster. \n", info_.serverList[myIndex_].executable.c_str());
    return 0;
  }
  else fclose(fp);

  switch (serverConfig.queue)
  {
    case QUEUE_PBS: printf("This run type is not yet implemented!\n"); return 0;break;
    case QUEUE_SGE: return runSGE(); break;
    case QUEUE_CUSTOM: return runCustom(); break;
    case QUEUE_MPICH: return runMPI(); break;
    case QUEUE_LAMMPI: return runMPI(); break;

    default: printf("Internal ERROR. TaskCtrl.run(), invalid queue type \n"); exit(0);
  }

  return 1;
}

///////////////////////////////////////////////////////////////////////////////
int TaskCtrl::loadServerInfo(string serverConfigFile)
{
  std::string path;

  mergePath(path, getNgPath(), serverConfigFile);
//  if (!serverConfig.loadInfo((char *)serverConfigFile.c_str()))
  if (!serverConfig.loadInfo((char *)path.c_str()))
  {
    printf("Can't load server's config file. \n");
    return 0;
  };

  return 1;
}

///////////////////////////////////////////////////////////////////////////////
char *TaskCtrl::genRunScriptName()
{
  static char name[255]; // = "ngRun.sh";
  char timeStr[20];
  time_t tm;

  time(&tm);
  sprintf(timeStr,"%i", tm);
 
  strcpy(name, info_.serverList[myIndex_].executable.c_str());
  strcat(name, "_ng");
  //strcat(name, timeStr);
  return name;
};

///////////////////////////////////////////////////////////////////////////////
int TaskCtrl::prepareRunScript(char *filenameSrc,char *filenameDst)
{
  int len;
  FILE *fpSource, *fpDest;
  char *buf, *pos;
  int i= 0;
  char tmpBuf[255];

  string envParams;
  string envKey;
  static char envParStr1[255];

  char saveDir[255];

  getcwd(saveDir, 255);
  chdir(ngPath_.c_str());
//  char saveDir1[255];
//  getcwd(saveDir1, 255);
//  printf("!!!!!!!!!! ngpath = %s,%s, %s\n", ngPath_.c_str(), saveDir1, filenameSrc);
  fpSource = fopen(filenameSrc,"rt");
  if (fpSource == NULL)
  {
    printf("ERROR. TaskCtrl().prepareRunScript. Can't open required run script %s for the queue\n", filenameSrc);
    chdir(saveDir);
    return 0;
  }

#ifdef _WIN32
  len = _filelength(_fileno(fpSource));
#else
  struct stat st;
  if (stat(filenameSrc, &st) ==-1)
  {
    printf ("Error in system call -'stat()'. Code  %d\n", errno);
    return 0;
  }
  len = st.st_size;
#endif

  chdir(saveDir);

  fpDest = fopen(filenameDst,"wt");
  if (fpDest == NULL)
  {
    printf("ERROR. TaskCtrl().prepareRunScript. Can't create new run script for the queue\n");
    return 0;
  }

  buf = (char *)malloc(len + 100);
  
  if (buf == NULL)
  {
    printf("ERROR. TaskCtrl().prepareRunScript. Can't allocate memory\n");
    return 0;
  }

  for (i=0; i < len; i++) buf[i] = 0;

  i = 0;
  buf[i] = fgetc(fpSource);

  while (buf[i] != EOF)
  {
    if ((pos = strstr(buf, NG_EXEC)) != NULL)
    {
      strcpy(pos, info_.serverList[myIndex_].executable.c_str());
      i += info_.serverList[myIndex_].executable.size() - strlen(NG_EXEC);
    };

    if ((pos = strstr(buf, NG_NP)) != NULL)
    {
      sprintf(tmpBuf,"%d", info_.serverList[myIndex_].nProcesses);  
      strcpy(pos, tmpBuf);
      i += strlen(tmpBuf) - strlen(NG_NP);
    };

    if ((pos = strstr(buf, NG_EXECARGS)) != NULL)
    {
      string str="";
      if (serverConfig.paramsPassMethod == PARAMS_ARG)
      {
        if (info_.serverList[myIndex_].execArgs.size() >0)
        {
          printf("ERROR. TaskCtrl. runMPI(). You can't combine ParamsPassMethod=PARAMS_ARG and passing arguments to the program through configuration file(execArgs=...). Try to set ParamsPassMethod = PARAMS_ENV in cluster.ini file in NumGRID home directory\n");
          return 0;
        };
        str += info_.projectFile;
        sprintf(tmpBuf,"%i", myIndex_);
        str +=" ";
        str += tmpBuf;
        str +=" ";
      }
    
      for (int j = 0; j < info_.serverList[myIndex_].execArgs.size(); j++)
      {
        str += info_.serverList[myIndex_].execArgs[j];
        str+=" ";
      }

      strcpy(pos, str.c_str());
      i += str.size() - strlen(NG_EXECARGS);
    };

    if ((pos = strstr(buf, (char *)serverConfig.addScript.c_str())) != NULL) // set new script name
    {
      strcpy(pos, genRunScriptName());
      i += strlen(genRunScriptName()) - serverConfig.addScript.size();
    };

    if ((pos = strstr(buf, NG_EXECENV)) != NULL)   // set environment variables
    {
      envParams = "";
      for (int j = 0; j < info_.serverList[myIndex_].execEnv.size(); j++)
      {
        putenv((char *)info_.serverList[myIndex_].execEnv[j].c_str());
      
        int pos1 = info_.serverList[myIndex_].execEnv[j].find_first_of("=");
        int pos2 = info_.serverList[myIndex_].execEnv[j].find_first_of(" ");
        if ((pos2 >0) && (pos2 < pos1)) pos1 = pos2;
      
        strncpy(envParStr1, info_.serverList[myIndex_].execEnv[j].c_str(),pos1);
        envParams+=envParStr1;
        envParams+=",";
      }
    
      envParams += "_GMPI_SPEC_FILE,_GMPI_SPEC_NUM";

      strcpy(pos, envParams.c_str());
      i += envParams.size() - strlen(NG_EXECENV);
    };

    buf[++i] = fgetc(fpSource);
  };

  fwrite(buf ,1 , i, fpDest);
  free(buf);
  fclose(fpSource); fclose(fpDest);
#ifdef _WIN32
#else
  chmod(filenameDst, S_IXUSR | S_IRUSR | S_IWUSR | S_IXUSR);
#endif
  return 1;
};

///////////////////////////////////////////////////////////////////////////////
int TaskCtrl::makeAllScripts()
{
  if (!prepareRunScript((char *)serverConfig.runScript.c_str(), (char *)NG_NEW_RUN_SH_NAME)) return 0;
//  if (!prepareRunScript("node.sh", genRunScriptName())) return 0;
  if (!prepareRunScript((char *)serverConfig.addScript.c_str(), genRunScriptName())) return 0;
    
  return 1;
};

