
int DEBUG_LEVEL  = 0 ; // Define this variable to see debug messages

#pragma warning(disable:4786)

#include <stdio.h>
#include <openssl/evp.h>

#ifdef _WIN32
  #include <windows.h>
  #include <io.h>
#else
  #include <pthread.h>
  #include <errno.h>
  #include <unistd.h>
  #include <sched.h>
  #include <wait.h>
#endif

#include "../Shared/projectinfo.h"
#include "../Shared/common.h"
#include "../Shared/net.h"

#include "main.h"
#include "client.h"


//////////////////////////////////////////////////////////////////////
// Check that symbol is an integer
int isNum(char symbol)
{
  char sym[] = {'0','1','2','3','4','5','6','7','8','9'};
  int j;

  for  (j=0; j < 10; j++) 
  {
    if (symbol == sym[j]) 
      break;
  }

  if (j < 10) return 1;
  else return 0;
};

//////////////////////////////////////////////////////////////////////
// delete duplicate integer values from the list
void deleteDuplicateNums(IntList *lst)
{
  IntList vars;
  int i, j;

  for (i=0; i < lst->size(); i++)
  {
    for (j=0; j < vars.size(); j++)
    {
      if ((*lst)[i] == vars[j]) break;
    };

    if (j == vars.size()) vars.push_back((*lst)[i]);
  }

  lst->assign(vars.begin(), vars.end());
  vars.clear();
};

//////////////////////////////////////////////////////////////////////
// Parse string of type 1,4,5,6-9,13,24-51,3   ...
IntList *parseNums(char **str)
{
  static IntList vars;
  char *posSpace, *curPos;
  char buf[255];
  bool n1Found, n2Found, mFound;
  int val1, val2;

  vars.clear();

  posSpace = strchr(*str, ' ');
  if (posSpace == NULL) posSpace = strchr(*str, 0);
  strncpy(buf, *str, posSpace- *str);
  buf[posSpace- *str] = 0;
  
  if (*posSpace == 0) *str = posSpace; else *str = posSpace+1;
  curPos = buf;

  n1Found =  n2Found = mFound = false; 
  for (int i =0; i < strlen(buf) + 1; i++)
  {
    if (isNum(buf[i]))
    {
      if (!mFound) n1Found = true; else n2Found = true;
      continue;
    }

    if (buf[i] == '-') 
    {
      if (mFound)
      {
        printf("ERROR. in position %d in string '%s'. Second '-' found. Should be integer value\n",i , buf);
        return NULL;
      };
      val1 = atoi(curPos);
      curPos = &buf[i] +1;
      mFound = true;
      continue;
    }

    if ((buf[i] == ',') || (buf[i] == 0))
    {
      if (strlen(buf) == 0) // no var params
      {
        return &vars;
      };
      if (!n1Found)
      {
        printf("ERROR. in position %d in string '%s' - should contain integer value before the separator\n",i , buf);
        return NULL;
      };
      if ((mFound) && (!n2Found))
      {
        printf("ERROR. in position %d in string '%s' - should contain integer value after '-'\n",i , buf);
        return NULL;
      };

      val2 = atoi(curPos);
      curPos = &buf[i] +1;

      if (mFound)
      {
        if (val1 >= val2)
        {
          printf("ERROR. in position %d in string '%s' - Value1 can't be greater than Value2 in Value1-Value2 pairs\n",i , buf);
          return NULL;
        };

        for (int j=val1; j < val2 +1; j++)
          vars.push_back(j);
      }
      else
      {
        vars.push_back(val2);
      };

      n1Found =  n2Found = mFound = false; 
      continue;
    };

    printf("ERROR. in position %d in string '%s' - Unknown symbol found\n",i , buf);
    return NULL;
  };

  return &vars;
};

//////////////////////////////////////////////////////////////////////
// Parse command line arguments
std::vector <CmdArgument> *parseArgs(int argc,char**argv, std::string &tName)
{
  int i, j;
  IntList *nums;
  CmdArgument arg;
  char *curPos;
  std::vector <CmdArgument> args;
  static std::vector <CmdArgument> argsNew;

  static std::string taskName = "";
  int buildFound, putdataFound, getdataFound, runFound, mpirunFound, killFound, allkillFound;

  buildFound = putdataFound = getdataFound = runFound = mpirunFound = killFound = allkillFound = -1;
  for (j = 1; j < argc; j++)
  {
    for (i=0; i < comTotal; i++)
    {
      if (!strcmpin(argv[j], comStr[i].name, strlen(comStr[i].name)) )
      {
        curPos=argv[j] + strlen(comStr[i].name);
        if ((nums = parseNums(&curPos)) == NULL) return NULL; 
        switch (comStr[i].sig)
        {
          case SIG_JOB_BUILD   : if (buildFound != -1) 
                                 {
                                   printf("ERROR.  More than one separate 'build' argument found\n");
                                   return NULL;
                                 } else buildFound = args.size();
                                 break;
          case SIG_JOB_PUTDATA : if (putdataFound != -1) 
                                 {
                                   printf("ERROR.  More than one separate 'putdata' argument found\n");
                                   return NULL;
                                 } else putdataFound = args.size();
                                 break;
          case SIG_JOB_GETDATA : if (getdataFound != -1) 
                                 {
                                   printf("ERROR.  More than one separate 'getdata' argument found\n");
                                   return NULL;
                                 } else getdataFound = args.size();
                                 break;
          case SIG_JOB_RUN     : if (runFound != -1) 
                                 {
                                   printf("ERROR.  More than one separate 'run' argument found\n");
                                   return NULL;
                                 } else runFound = args.size();
                                 break;
                                 if (nums->size() != 0)
                                 {
                                   printf("ERROR.  '%s' - 'Run' argument should take no integer parameters. This command works on all NumGRID resources\n", argv[j]);
                                   return NULL;
                                 }
                                 break;
          case SIG_JOB_RUN_MPI : if (mpirunFound != -1) 
                                 {
                                   printf("ERROR.  More than one separate 'mpirun' argument found\n");
                                   return NULL;
                                 } else mpirunFound = args.size();
                                 break;
                                 if (nums->size() != 0)
                                 {
                                   printf("ERROR.  '%s' - 'Mpirun' argument should take no integer parameters. This command works on all NumGRID resources\n", argv[j]);
                                   return NULL;
                                 }
                                 break;
          case SIG_JOB_KILL    : if (killFound != -1) 
                                 {
                                   printf("ERROR.  More than one separate 'delete' argument found\n");
                                   return NULL;
                                 } else killFound = args.size();
                                 if (nums->size() != 0)
                                 {
                                   printf("ERROR.  '%s' - 'delete' argument should take no integer parameters. This command works on all NumGRID resources\n", argv[j]);
                                   return NULL;
                                 }
                                 break;
          case SIG_JOB_ALLKILL : if (allkillFound != -1) 
                                 {
                                   printf("ERROR.  More than one separate 'killall' argument found\n");
                                   return NULL;
                                 } else allkillFound = args.size();
                                 if (nums->size() != 0)
                                 {
                                   printf("ERROR.  '%s' - 'killall' argument should take no integer parameters. This command works on all NumGRID resources\n", argv[j]);
                                   return NULL;
                                 }
                                 break;
        };

        arg.sig = comStr[i].sig;
        arg.vars.assign(nums->begin(), nums->end());
        deleteDuplicateNums(&arg.vars);
        args.push_back(arg);
        break;
      };
    }

    if (i == comTotal)
    {
      if (taskName == "")
      {
        taskName = argv[j];
      }
      else
      {
        printf("ERROR.  '%s' - Unknown command line argument found\n", argv[j]);
        return NULL;
      }
    }
  }
  // Sort args in such order (putdata, build, run, getdata,...)
  if (putdataFound != -1) argsNew.push_back(args[putdataFound]);
  if (buildFound != -1) argsNew.push_back(args[buildFound]);
  if (runFound != -1) argsNew.push_back(args[runFound]);
  if (mpirunFound != -1) argsNew.push_back(args[mpirunFound]);
  if (getdataFound != -1) argsNew.push_back(args[getdataFound]);
  if (killFound != -1) argsNew.push_back(args[killFound]);
  if (allkillFound != -1) argsNew.push_back(args[allkillFound]);

  args.clear();
  //for (i = 0; i < argsNew.size(); i++) printf("%d\n", argsNew[i].sig);
  tName = taskName;
  return &argsNew;
};

//////////////////////////////////////////////////////////////////////
void showHelp()
{
  printf("Syntax: ngclient <specFile> <key1[a1[-ai]] [key2[b1[-bj]]]...[keyN[c1[-ck]]]\n");
  printf("Keys:\n");
  printf("-putdata   .Upload input files to cluster(s) [a1-[an]]\n");
  printf("-getdata   .Download output files from cluster(s) [a1-[an]]\n");
  printf("-build     .Build source files on cluster(s) [a1-[an]]\n");
  printf("-run       .Run the job on all clusters\n");
  printf("-mpirun    .Run mpi job on all clusters\n");
  printf("-killall   .Kill all NumGRID processes and task on all clusters\n");
  printf("Examples:\n");
  printf("ngclient dateTest.ng -run\n");
  printf("ngclient mpiTEstRing.ng - build0 -mpirun\n");
  printf("ngclient GalaxyTEst.ng -putdata1 -build0-2 -mpirun -getdata0\n");
  printf("\n");
  printf("\n");
 
}

//////////////////////////////////////////////////////////////////////
int main(int argc,char**argv)
{
  std::vector <CmdArgument> *args;
  ProjectInfo project;
  int i;
  Client *cl;
  char username[255], pass[255];

#ifdef _WIN32
  HANDLE *th;
#else
  pthread_t *th;
#endif

  std::string pName="";

  //args = parseArgs("-build1,1-11,5,2-3 -putdata5 -run");
  //args = parseArgs("-build parra.ng -putdata -getdata");
  //args = parseArgs("-build  -putdata -getdata");
  printf("NumGRID client. Version 1.0. (c) SSD, 2007\n");
  
  args = parseArgs(argc, argv, pName);
  if (args == NULL) 
  {
    showHelp();
    return 1;
  }

  if (!initNetwork()) return 1;

  if (pName == "")
  {
    showHelp();
    printf("ERROR. You must specify a project file in commnad line arguments\n");
    return 0;
  };

  if (!project.LoadInfo((char *)pName.c_str()))
    return 1;

#ifdef _WIN32
  th=(HANDLE*)malloc(sizeof(HANDLE)*project.serverList.size());
  for (i =0; i < project.serverList.size(); i++) th[i] = NULL;
#else
  th=(pthread_t*)malloc(sizeof(pthread_t)*project.serverList.size());
  for (i =0; i < project.serverList.size(); i++) th[i] = 0;
#endif

  // create all client threads
  for (i=0; i < project.serverList.size(); i++)
  {
    // check if we don't need to start some threads because they will have nothing to do
    bool foundNum = false;
    for (int a =0; a < args->size(); a++)
    {
      if ((*args)[a].vars.size() == 0) foundNum = true;
      else
      {  //not all servers are operated by this command
        for (int j =0; j < (*args)[a].vars.size(); j++)
          if (i == (*args)[a].vars[j]) foundNum = true;
      }
    }
    if (!foundNum) continue;

    // Get username & password if absent
    if (project.serverList[i].username.empty())
    {
     printf("Please enter your username for cluster '%s':", project.serverList[i].blockName.c_str());
     gets(username);
     project.serverList[i].username = username;
    }
    if (project.serverList[i].password.empty())
    {
     std::string tmp;
     tmp = "Please enter your password for cluster '";
     tmp += project.serverList[i].blockName;
     tmp += "':";
     EVP_read_pw_string(pass,255,tmp.c_str(),0);
     project.serverList[i].password = pass;
    }
    // ----
    
    cl = new Client;
    cl->project = project;
    cl->myPos = i;
    cl->args =  args;
    cl->projectFilename = pName;

    #ifdef _WIN32
    DWORD id;
    th[i] = CreateThread(NULL, 0, clientThread, (void *)cl, 0, &id);
    if (th[i]==NULL)
    {
      printf("ERROR. Can't start new thread: %s\n",strerror(errno));
    }
    #else

    if(pthread_create(&th[i],0,(void*(*)(void*))clientThread, (void *)cl))
    {
      printf("ERROR. Can't start new thread: %s\n",strerror(errno));
    }
    #endif
  };

  // wait all client threads
  for (i=0; i < project.serverList.size(); i++)
  {
    #ifdef _WIN32
    if (th[i] == NULL) continue;

    DWORD code;
    do 
    {
      GetExitCodeThread(th[i], &code);
      Sleep(100);
    }
    while (code == STILL_ACTIVE);

    #else
    if (th[i] == 0)  continue;

    int p,st;
    void *ptr=0;
    pthread_join(th[i],&ptr);

    #endif
  }

  closeNetwork();
  return 0;
};
