
#define USE_PAM
//#define USE_SHADOW
//#define USE_SIMPLE

#ifdef USE_SHADOW 
  #define _XOPEN_SOURCE
  #include <shadow.h>
  #undef _XOPEN_SOURCE
#endif

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>


#ifndef _WIN32
  #include <unistd.h>
  #include <sched.h>
  #include <pthread.h>
  #include <pwd.h>
  #include <sched.h>
  #include <wait.h>
  #include <sys/time.h>
  #include <time.h>

  #ifdef USE_PAM 
    #include <security/pam_appl.h>
  #endif
#endif

#ifdef USE_PAM
//////////////////////////////////////////////////////////////////////
int numgrid_conv(int n,const struct pam_message **msgm,struct pam_response **response,void *appdata_ptr)
{
  int i;
  char *user,*pass;
  struct pam_response *resp1;

  struct APPDATA
  {
    char *user,*pass;
  }
  *appdata = (struct APPDATA*) appdata_ptr;

  if(n<=0||n>PAM_MAX_NUM_MSG) return PAM_BUF_ERR;

  resp1 = (struct pam_response*)malloc(n*sizeof(struct pam_response));
  user = strdup(appdata->user);
  pass = strdup(appdata->pass);

  if(!(pass&&user)) return PAM_CONV_ERR;

  for(i=0;i<n;i++)
  {
    resp1[i].resp=0;
    resp1[i].resp_retcode=0;
    switch(msgm[i]->msg_style)
    {
      case PAM_PROMPT_ECHO_OFF:resp1[i].resp=strdup(pass);
                               resp1[i].resp_retcode=0;
                               break;
      case PAM_PROMPT_ECHO_ON://resp1[i].resp=user;
      resp1[i].resp=strdup(pass);
                              resp1[i].resp_retcode=0;
                              break;
      default:*response=0;
              return PAM_CONV_ERR;
              break;
    }
  }

  *response=resp1;
  return PAM_SUCCESS;
}
#endif

//////////////////////////////////////////////////////////////////////
int authenticate(char *user, char*pass)
{
#ifndef _WIN32

  struct passwd *pwd;
  pwd=getpwnam(user);
  if(!pwd)
  {
    printf("ERROR. Can't authenticate user %s: No such user\n",user);
    return 0;
  }

#ifdef USE_PAM
  struct {char *user,*pass;}appdata={user,pass};
  pam_handle_t *ph;
  struct pam_conv pc;
  int rd;

  pc.conv=numgrid_conv;
  pc.appdata_ptr=&appdata;

  rd=pam_start("login",user,&pc,&ph);
  if(rd != PAM_SUCCESS)
  {
    printf("ERROR. Can't start pam module %s\n","su");
    printf("%s\n",pam_strerror(ph,rd));
    return 0;
  }

  rd=pam_authenticate(ph,0);
  if(rd!=PAM_SUCCESS)
  {
    printf("ERROR. Can't authenticate user %s\n",user);
    printf("%s\n",pam_strerror(ph,rd));
    pam_end(ph, rd);
    return 0;
  }
  
  pam_end(ph, rd);
  return pwd->pw_uid;

#elif defined USE_SHADOW 
  struct spwd *spwd;
  char *epass;

  if(geteuid() != 0)
  {
    printf("ERROR. You can't use SHADOW authentication because you are not ROOT\n");
    return 0;
  }

  spwd=getspnam(user);
  if (spwd == NULL)
  {
    printf("ERROR. Can't authenticate user %s: No such user\n",user);
    return 0;
  }
  
  epass=crypt(pass,spwd->sp_pwdp);
//  printf("user= %s, pass = %s, epass = %s, %s\n", user, pass, epass,spwd->sp_pwdp);
  if ((strcmp(epass,spwd->sp_pwdp) != 0) || (strlen(epass) != strlen(spwd->sp_pwdp)))
  {
    printf("ERROR. Authentication failed\n");
    return 0;
  }
  return pwd->pw_uid;

#elif defined USE_SIMPLE
  struct passwd *pwd;
  struct passwd *pwd;

  if(geteuid() == 0)
  {
    printf("ERROR. You can't use SIMPLE authentication because you are ROOT\n");
    return 0;
  }

  pwd=getpwuid(geteuid());
  if (strcmp(pwd->pw_name, user) != 0)
  {
//    printf("%s %s, %s %s\n", pwd->pw_name, pwd->pw_passwd, user, pass);
    printf("ERROR. Can't authenticate user %s: Login does not match\n",user);
    return 0;
  }
  
#else
  printf("WARNING!!!: You don't use any authority system\n");
  printf("WARNING!!!: Your computer can be accessed by everyone\n");
#endif

  return pwd->pw_uid;
#endif
// Windows
  return 1;
}

