
#pragma warning(disable:4786)

#ifdef _WIN32
  #include <windows.h>
#else  
 #include <stdio.h>
 #include <errno.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <string.h>
 #include <stdlib.h>
#endif

#include "job.h"

extern volatile bool needKillAll;

//////////////////////////////////////////////////////////////////////
int Job::execBlocked(char *filename, char *args[], char *fDir)
{
  char dir[MAX_DIRECTORY_LENGTH];
  char fName[20];
  FILE *fp;
  std::string err;
  
  getcwd(dir, MAX_DIRECTORY_LENGTH);
  chdir(fDir);
  if (!spawn(filename, args, fName, id))
  {
    err = "ERROR. Can't run remote file. May be there is no such file on remote computer or filename/path are wrong\nFilename =  ";
    err+=filename; err+="\n";
    writeMessage(es, (char *)err.c_str(), SIG_ERROR);
    chdir(dir);
    return 0;
  };
  fp = fopen(fName, "rt");
  chdir(dir);

  // send realtime output  
  int ret, st;
  char str[255];
  if (fp == NULL) 
  {
    writeMessage(es, "Can't open file with output\n", SIG_ERROR);
    return 0;
  }
  
  do
  {
    ret = waitpid(id, &st, WNOHANG);
    if (( ret < 0) && (errno == ECHILD)) 
      break;
      
    if( ret < 0)   
    {
      kill(id,SIGKILL);
      printf("ERROR. exec. waitpid: errno = %d \n", errno);
      return 0;
    }
    usleep(100);
    fflush(fp);
    fsync(fileno(fp));
    if (fgets(str, 255, fp) != NULL) 
      if (!writeMessage(es, str, SIG_CONT)) 
      {
        fclose(fp);
        return 0;
      };
  }
  while (!((ret == id) && WIFEXITED(st)));
  id = -1;
  usleep(10);
  while (fgets(str, 255, fp) != NULL) writeMessage(es, str, SIG_CONT);
  
  fclose(fp);
  //writeMessage(es, "Build successfully completed\n", SIG_OK);
  return 1;
};

//////////////////////////////////////////////////////////////////////
int Job::Build()
{
  char *args[5];
  std::string path;
  FILE *fp;
  
  debug(1, "Building new job...\n");

  path = project.serverList[myPos].workDir;
  if (!readFile(es, "", project.serverList[myPos].workDir))
  {
    writeMessage(es, "Can't receive makefile\n", SIG_ERROR);
    return 0;
  }
  if (!readFileList(es, project.serverList[myPos].workDir))
  {
    writeMessage(es, "Can't receive source files\n", SIG_ERROR);
    return 0;
  }

  // make
  args[0] = "make";
  args[1] = "-f";
  args[2] = (char *)project.serverList[myPos].makefile.c_str();
  args[3] = NULL;
  
  if (!execBlocked("make",args, (char *)project.serverList[myPos].workDir.c_str())) 
    return 0;
  
  // check
  if ((fp = fopen(mergePath(path,project.serverList[myPos].workDir,
            project.serverList[myPos].executable).c_str(),"r")) == NULL) 
  {
    std::string out = "No executable found after Build operation. Build may be incomplete. Searched for ";
    out += path; out+="\n";
    writeMessage(es, (char *)out.c_str(), SIG_OK);
  }    
  else
  {
    fclose(fp);
    writeMessage(es, "Build successfully completed\n", SIG_OK);
  }    
  debug(1, "Build completed\n");
  return 1;
};

//////////////////////////////////////////////////////////////////////
int Job::Putdata()
{
  debug(1, "Uploading data files for the job...\n");
  if (!readFileList(es, project.serverList[myPos].workDir))
  {
    writeMessage(es, "Can't receive data files\n", SIG_ERROR);
    return 0;
  }

  writeMessage(es, "Data files upload completed\n", SIG_OK);
  debug(1, "Data files upload completed\n");
  return 1;
};

//////////////////////////////////////////////////////////////////////
int Job::Getdata()
{
  debug(1, "Downloading job result files...\n");
  if (!writeFileList(es, project.serverList[myPos].outputFiles, project.serverList[myPos].workDir)) 
  {
    writeMessage(es, "Can't send result files\n", SIG_ERROR);
    return 0;
  }

  writeMessage(es, "Job result files download completed\n", SIG_OK);
  debug(1, "Job result files download completed\n");
  return 1;
};

//////////////////////////////////////////////////////////////////////
// run using Jobmanager
int Job::Run()
{
  char *args[4];
  char fName[20];
  char dir[MAX_DIRECTORY_LENGTH];
  int hn;
  std::string jmanPath;
  int i;

  mergePath(jmanPath,getNgPath(), "jobmanager");

  args[0] = new char[20];  
  args[1] = new char[project.projectFile.size()];
  args[2] = new char[10];
   
  strcpy(args[0], "jobmanager");
  strcpy(args[1], project.projectFile.c_str());
  sprintf(args[2],"%d", myPos);
  
  args[3] = NULL;
//  TaskCtrl task;
  debug(1, "Running a new grid-mpi job...\n");
  //task.init(project, myPos, getNgPath());
//  task.loadServerInfo(SERVER_CFG_FILE)
printf("---Zapuskayu %s %s %s, myPos = %d\n", args[0], args[1], args[2], myPos);
  if (!execBlocked((char *)jmanPath.c_str(), args, (char *)project.serverList[myPos].workDir.c_str())) 
    return 0;

  for (i =0; i < 3; i++) delete args[i];

  writeMessage(es, "Job was successfully completed\n", SIG_OK);
  debug(1, "Job was successfully completed\n");
  return 1;
};

//////////////////////////////////////////////////////////////////////
// run without jobmanager (non MPI job)
int Job::RunSimple()
{
  char *args[255];
  char fName[20];
  char dir[MAX_DIRECTORY_LENGTH];
  int hn;
  FILE *fp;
  int i;
  std::string path;

  debug(1, "Running a new job...\n");

  args[0] = new char[project.serverList[myPos].executable.size()];
  strcpy(args[0], (char *)project.serverList[myPos].executable.c_str());

  for (i=1; i < project.serverList[myPos].execArgs.size() + 1; i++)
  {
    args[i] = new char[project.serverList[myPos].execArgs[i].size()];
    strcpy(args[i], project.serverList[myPos].execArgs[i].c_str());
  }
    
  args[i] = NULL;
  // run without jobmanager (non MPI job)

  //set +x mod before execution
  mergePath(path,project.serverList[myPos].workDir, project.serverList[myPos].executable);
  chmod(path.c_str(),S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
  
  if (!execBlocked((char *)project.serverList[myPos].executable.c_str(),
                   args, (char *)project.serverList[myPos].workDir.c_str()))
  {	 
//    writeMessage(es, "ERROR. Can't run remote job. May be there is no such file on remote computer or filename/path are wrong\n", SIG_ERROR);
    return 0;
  }
  
  for (i=0; i < project.serverList[myPos].execArgs.size() + 1; i++) delete args[i];

  writeMessage(es, "Job successfully completed\n", SIG_OK);
  debug(1, "Job successfully completed\n");
  return 1;
};

//////////////////////////////////////////////////////////////////////
int Job::Kill()
{
  debug(1, "Killing the job...\n");

  writeMessage(es, "The job was successfully killed\n", SIG_OK);
  debug(1, "The job was successfully killed\n");
  return 1;
};

//////////////////////////////////////////////////////////////////////
int Job::AllKill()
{
  std::string com;
  debug(1, "Killing all jobs for this user...\n");
  // kill processes
  needKillAll = true;
  
  // kill in queues
  // mvs
  com = "mkill -u ";
  com += project.serverList[myPos].username;
  system(com.c_str());

  com = "mqdel -u ";
  com += project.serverList[myPos].username;
  system(com.c_str());

  // sge
  com = "qdel -u ";
  com += project.serverList[myPos].username;
  system(com.c_str());
  
  writeMessage(es, "All jobs were successfully killed\n", SIG_OK);
  // jobmanager
  system("killall jobmanager");
  system("killall mpirun");
    
  debug(1, "All jobs were successfully killed\n");
  return 1;
};

//////////////////////////////////////////////////////////////////////
void Job::Terminate()
{
  closeSocket(((_ENCSOCK *)es)->soc);
  ((_ENCSOCK *)es)->soc = -1;
  debug(1,"Connection closed. \n");
#ifdef _WIN32
  ExitThread(0);
#else
  pthread_exit(0);  
#endif
};

//////////////////////////////////////////////////////////////////////
