
int DEBUG_LEVEL  = 4 ; // Define this variable to see debug messages

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include <openssl/rsa.h>
#include <openssl/ssl.h>
#include <openssl/rand.h>

#ifdef _WIN32
  #include <io.h>
#else
  #include <errno.h>
  #include <signal.h>
#endif

#include "auth.h"
#include "../Shared/net.h"
#include "../Shared/esock.h"
#include "job.h"
//#include "server.h"

#ifdef _WIN32
  DWORD WINAPI serverThread(LPVOID ptr);
#else
  void *serverThread(void *ptr);
#endif

JobList jobs;
volatile bool needKillAll = false;

//////////////////////////////////////////////////////////////////////
int getNewConnection(int servSoc)
{
  RSA *rsa;
  int cSoc;
  _ENCSOCK *_es;
  ENCSOCK es;
  Job job;
  Packet pack;
  char buf[1024],user[128],pass[128];

  srandom(12);
  rsa=RSA_generate_key(2048,RSA_F4,0,0);

  cSoc = acceptConnection(servSoc);
  debug(1,"Got connection from a client\n");
  if(cSoc == -1) return 1;

  es = esockAccept(cSoc,rsa);
  _es=(_ENCSOCK *)es;

  if(!es)
  {
    printf("ERROR. Can't establish security context\n");
    closeSocket(cSoc);
    return 0;
  }
  else debug(1,"Security context established\n");

  pack.data = buf;
  if(!esockReadPacket(es,&pack)) 
  {
    printf("ERROR. Can't read username and password\n");
    writeMessage(es, "SERVER ERROR. Can't read username and password\n", SIG_ERROR);
    closeSocket(cSoc);
    return 0;
  }
  debug(1,"Got username and password\n");

  pack.data[pack.header.size]=0;
  if(sscanf(pack.data,"%128[^:]:%128s",user,pass)!=2)
  {
    printf("ERROR. Can't parse username and password. Expected <username>:<password>\n");
    writeMessage(es, "SERVER ERROR. Can't parse username and password. Expected <username>:<password>\n", SIG_ERROR);
    closeSocket(cSoc);
    return 0;
  }

  if((job.userId = authenticate(user,pass)) == 0)
  {
    printf("ERROR. Authentication failed\n");
    writeMessage(es, "SERVER ERROR. Authentication failed\n", SIG_ERROR);
    closeSocket(cSoc);
    return 0;
  }
  else writeMessage(es, " ", SIG_OK);
  debug(1,"User successfully authentificated\n");

  job.es = es;
  jobs.push_back(job);

  return 1;
};

//////////////////////////////////////////////////////////////////////
int main(int argc,char**argv)
{
  int servSoc;
  fd_set socs;
  timeval selTime;
#ifdef _WIN32
  HANDLE th;
#else
  pthread_t th;
#endif

//  struct passwd *pwd;

  debug(1,"NumGRID server. Version 1.0. (c) SSD, 2007\n");
  if (getNgPath() == "") return 1;
  
  initNetwork();

  servSoc = makeServerSocket(DEFAULT_SERVER_PORT);
  if(servSoc == -1) return 1;

  while (true)
  {
    FD_ZERO(&socs);
    FD_SET(servSoc, &socs);

    selTime.tv_sec = 0;
    selTime.tv_usec = 500;

    select(65535, &socs, NULL, NULL, &selTime);
//    select(65535, &socs, NULL, NULL, NULL);
    if (FD_ISSET (servSoc, &socs)) // new connection
    {
      if (getNewConnection(servSoc))
      {
       #ifdef _WIN32
         DWORD id;
         th = CreateThread(NULL, 0, serverThread, (void *)&jobs.back(), 0, &id);
         if (th==NULL)
         {
           printf("ERROR. Can't start new thread: %s\n",strerror(errno));
         }
       #else
         if(pthread_create(&th,0,(void*(*)(void*))serverThread, (void *)&jobs.back()))
         {
           printf("ERROR. Can't start new thread: %s\n",strerror(errno));
         }
       #endif
      }
    }

    JobList::iterator ji; // remove closed links

    if (needKillAll)
    {
      needKillAll = false;
      for (ji = jobs.begin(); ji != jobs.end(); ji++)
      {
        if (jobs.empty()) break;
	if (ji->id != -1) kill(ji->id, SIGKILL);
        closeSocket(((_ENCSOCK *)ji->es)->soc);
        ((_ENCSOCK *)ji->es)->soc = -1;
      }
    };
    for (ji = jobs.begin(); ji != jobs.end(); ji++)
    {
      if (jobs.empty()) break;
      if (((_ENCSOCK *)ji->es)->soc == -1) jobs.erase(ji);
    }
  }

  return 0;
};
