
#pragma warning(disable:4786)

#ifdef _WIN32
  #include <io.h>
  #include <direct.h>
  #include <windows.h>
#else
  #include <unistd.h>
  #include <errno.h>
  #include <pwd.h>
  #include <sys/types.h>
  #include <sys/stat.h>
#endif

#include "../Shared/projectinfo.h"
#include "../Shared/common.h"
//#include "server.h"
#include "job.h"

//////////////////////////////////////////////////////////////////////
#ifdef _WIN32
DWORD WINAPI serverThread(LPVOID ptr)
#else
void *serverThread(void *ptr)
#endif
{
  char *mes = NULL;
  int len;
  int sig;
  Job job = *(Job *)ptr;
  char pName[15];
  std::string path, path1;
  struct passwd *pwd;

  if(setuid(job.userId) == -1)
  {
    debug(1,"ERROR. Can't switch to user ID = %d.\n", job.userId);
    job.Terminate();
  }

  strcpy(pName, "ng_specXXXXXX");
  if (mktemp(pName) == NULL)
  {
    printf("ERROR. Can't create temporary filename for output.\n");
    job.Terminate();
  };
  
  pwd = getpwuid(job.userId);
  if (pwd == NULL)
  {
    debug(1,"ERROR. Can't find user id ID = %d. \n", job.userId);
    job.Terminate();
  }
  
  if (!readFile(job.es, pName, pwd->pw_dir))
  {
    debug(1,"ERROR. Can't read file with a project description. \n");
    writeMessage(job.es, "ERROR. Can't read file with a project description.\n", SIG_ERROR);
    job.Terminate();
  };

  if (!readMessage(job.es, &mes, len, sig))
  {
    debug(1,"ERROR. Can't read a message.\n");
    writeMessage(job.es, "ERROR. Can't read a message.\n", SIG_ERROR);
    
    job.Terminate();
  };
  free(mes);

  mergePath(path, pwd->pw_dir, pName);
  if (!job.project.LoadInfo((char *)path.c_str())) 
  {
    writeMessage(job.es, "ERROR. Can't load project description. Syntax error in the file.\n", SIG_ERROR);
    job.Terminate();
  }
  job.myPos = sig;

  if (job.project.serverList[job.myPos].workDir.size() == 0 ) // if worker="" set workdir = $HOME
    job.project.serverList[job.myPos].workDir = pwd->pw_dir;

  mergePath(path, (char *)job.project.serverList[job.myPos].workDir.c_str(), pName);
  mergePath(path1, pwd->pw_dir, pName);
#ifdef _WIN32
  mkdir(job.project.serverList[job.myPos].workDir.c_str());
#else
  int res = mkdir(job.project.serverList[job.myPos].workDir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH);
  // check that workdir is correct. if not set it to $HOME
  if ((res == -1) && (errno != EEXIST)) job.project.serverList[job.myPos].workDir = pwd->pw_dir;
#endif  
  if (rename(path1.c_str(), path.c_str()) == -1)
  {
    job.project.serverList[job.myPos].workDir = pwd->pw_dir;
    mergePath(path, (char *)job.project.serverList[job.myPos].workDir.c_str(), pName);
    rename(path1.c_str(), path.c_str());
  };
  
  job.project.projectFile = path;
  
  while (readMessage(job.es, &mes, len, sig))
  {
    switch (sig)
    {
      case SIG_OK: break;
      case SIG_ERROR      : printf("Action received\n"); break;
      case SIG_JOB_BUILD  : if (!job.Build()) job.Terminate(); break;
      case SIG_JOB_RUN    : if (!job.RunSimple()) job.Terminate(); break;
      case SIG_JOB_RUN_MPI: if (!job.Run()) job.Terminate(); break;
      case SIG_JOB_KILL   : if (!job.Kill()) job.Terminate(); break;
      case SIG_JOB_ALLKILL: if (!job.AllKill()) job.Terminate(); break;
      case SIG_JOB_PUTDATA: if (!job.Putdata()) job.Terminate(); break;
      case SIG_JOB_GETDATA: if (!job.Getdata()) job.Terminate(); break;
      case SIG_GET_OUTPUT : printf("Action SIG_GET_OUTPUT received\n"); break;
      case SIG_GET_STATE  : printf("Action SIG_GET_STATE received\n"); break;

      default: printf("Unknown action signal received\n"); return 0;
    };

    free(mes);
  }

  job.Terminate();
  return 0;
}
